from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any

import altair as alt
import numpy as np
import polars as pl

from ..logging import logger


@dataclass
class BarMeta:
    """
    Metadata for chart visualization.
    """

    data: list[dict[str, Any]]
    x_label: str
    y_label: str
    title: str

    def to_dict(self) -> dict[str, Any]:
        return {
            "data": self.data,
            "x_label": self.x_label,
            "y_label": self.y_label,
            "title": self.title,
        }


@dataclass
class HeatmapMeta:
    """
    Metadata for heatmap visualization.
    """

    x: list[str]
    y: list[str]
    x_label: str
    y_label: str
    title: str
    matrix: np.ndarray
    save_path: Path

    def __post_init__(self):
        if not isinstance(self.matrix, np.ndarray):
            raise ValueError("Matrix must be a numpy ndarray")
        if not self.save_path.parent.exists():
            self.save_path.parent.mkdir(parents=True, exist_ok=True)

    def to_dict(self) -> dict[str, Any]:
        return {
            "x": self.x,
            "y": self.y,
            "x_label": self.x_label,
            "y_label": self.y_label,
            "title": self.title,
            "matrix": self.matrix.tolist(),
            "save_path": str(self.save_path),
        }


@dataclass
class BarConfig:
    """
    Configuration for common distribution visualization.
    """

    x_label: str
    title: str
    filename: str
    output_csv: bool

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)


@dataclass
class HeatmapConfig:
    """
    Configuration for heatmap chart visualization.
    """

    y_label: str
    title: str
    filename: str

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)


@dataclass
class VisInjectionsConfig:
    """
    Configuration for visualization of injections.
    """

    bar_configs: dict[str, BarConfig]
    heatmap_configs: dict[str, HeatmapConfig]

    def to_dict(self) -> dict[str, Any]:
        return {
            "bar_configs": {key: config.to_dict() for key, config in self.bar_configs.items()},
            "heatmap_configs": {key: config.to_dict() for key, config in self.heatmap_configs.items()},
        }


def NewVisInjectionsConfig() -> VisInjectionsConfig:
    return VisInjectionsConfig(
        bar_configs={
            "faults": BarConfig(
                x_label="Fault Type",
                title="Fault Distribution",
                filename="faults",
                output_csv=True,
            ),
            "services": BarConfig(
                x_label="Service Name",
                title="Service Distribution",
                filename="services",
                output_csv=True,
            ),
        },
        heatmap_configs={
            "fault_services": HeatmapConfig(
                y_label="Service Name",
                title="Fault-Service Count Distribution",
                filename="fault_services",
            ),
            "fault_pair_attribute_coverages": HeatmapConfig(
                y_label="Pair Name",
                title="Fault-Pair Attribute Coverage Distribution",
                filename="fault_pair_attribute_coverages",
            ),
            "fault_service_attribute_coverages": HeatmapConfig(
                y_label="Service Name",
                title="Fault-Service Attribute Coverage Distribution",
                filename="fault_service_attribute_coverages",
            ),
        },
    )


def plot_bar(meta: BarMeta) -> alt.Chart:
    df = pl.DataFrame(meta.data)
    base = alt.Chart(data=df.to_pandas())

    degree_selection = alt.selection_point(fields=["degree"], bind="legend")
    degree_color = alt.Color(
        "degree",
        title="Degree",
        scale=alt.Scale(
            domain=["absolute_anomaly", "may_anomaly", "no_anomaly"],
            range=["#e74c3c", "#f39c12", "#2ecc71"],
        ),
    )

    degree_opacity = alt.when(degree_selection).then(alt.value(0.7)).otherwise(alt.value(0.1))

    tooltip_fields = ["degree", "name", "count"]

    if "chaos_type" in df.columns:
        tooltip_fields.append("chaos_type:N")

    bar = (
        base.mark_bar(stroke="navy", strokeWidth=1, opacity=0.7)
        .encode(
            x=alt.X(
                "name",
                title=meta.x_label,
                axis=alt.Axis(labelAngle=-45),
                sort="-y",
            ),
            y=alt.Y("count", title="Injection Count"),
            color=degree_color,
            opacity=degree_opacity,
            xOffset="degree",
            tooltip=tooltip_fields,
        )
        .properties(width=850, height=300, title=meta.title)
        .add_params(degree_selection)
    )

    return bar


def plot_heatmap(meta: HeatmapMeta) -> None:
    pass


class VisDatapacks:
    BAR_DATA_KEY = ["degree", "name", "count"]

    def __init__(self, config: VisInjectionsConfig, distributions_dict: dict[str, dict[str, Any]], metrics: list[str]):
        self.config = config
        self.distributions_dict = distributions_dict
        self.metrics = metrics

    def display_bar(self, bar_data: list[dict[str, Any]], key: str) -> alt.Chart | None:
        """
        Display bar charts for faults, services distribution.
        """

        for item in bar_data:
            if not all(k in item for k in self.BAR_DATA_KEY):
                logger.error(f"Bar data item {item} is missing required keys: {self.BAR_DATA_KEY}")
                return None

        if key not in self.config.bar_configs:
            logger.error(f"Key {key} not found in bar_configs")
            return None

        bar_config = self.config.bar_configs[key]

        return plot_bar(
            BarMeta(
                data=bar_data,
                x_label=bar_config.x_label,
                y_label="Injection Count",
                title=bar_config.title,
            )
        )

    def get_bar_data(self, key: str) -> list[dict[str, Any]]:
        combined_data = []

        for degree, distributions in self.distributions_dict.items():
            data: dict[str, int] = distributions.get(key, {})
            if not data:
                continue

            for name, count in data.items():
                combined_data.append(
                    {
                        "degree": degree,
                        "name": name,
                        "count": count,
                    }
                )

        return combined_data



import os

import polars as pl
from rcabench.openapi import (
    ApiClient,
    AuthenticationApi,
    Configuration,
    DatasetsApi,
    DtoAlgorithmDatapackReq,
    DtoAlgorithmDatapackResp,
    DtoAlgorithmDatasetReq,
    DtoAlgorithmDatasetResp,
    DtoDatapackEvaluationBatchReq,
    DtoDatasetEvaluationBatchReq,
    DtoGranularityRecord,
    DtoInjectionV2Response,
    DtoInjectionV2SearchReq,
    EvaluationApi,
    InjectionsApi,
    ProjectsApi,
)
from rcabench.openapi.models.dto_login_request import DtoLoginRequest
from rcabench.rcabench import RCABenchSDK

from ..config import get_config
from ..logging import logger


def get_rcabench_sdk(*, base_url: str | None = None) -> RCABenchSDK:
    if base_url is None:
        base_url = get_config().base_url

    return RCABenchSDK(base_url=base_url)


def get_rcabench_openapi_client(*, base_url: str | None = None) -> ApiClient:
    if base_url is None:
        base_url = get_config().base_url

    return ApiClient(configuration=Configuration(host=base_url))


class RCABenchClient:
    """
    Usage:
    with RCABenchClient() as api_client:
        container_api = rcabench.openapi.ContainersApi(api_client)
        containers = container_api.api_v2_containers_get()
        print(f"Containers: {containers.data}")
    """

    _instances = {}
    _sessions = {}

    def __new__(cls, base_url: str | None = None, username: str | None = None, password: str | None = None):
        # Parse actual configuration values
        actual_base_url = (
            base_url
            or os.getenv("RCABENCH_BASE_URL")
            or get_config(env_mode=os.environ.get("ENV_MODE", "dev")).base_url
        )
        actual_username = username or os.getenv("RCABENCH_USERNAME")

        # Use (base_url, username) as unique identifier
        instance_key = (actual_base_url, actual_username)

        if instance_key not in cls._instances:
            instance = super().__new__(cls)
            cls._instances[instance_key] = instance
            instance._initialized = False

        return cls._instances[instance_key]

    def __init__(self, base_url: str | None = None, username: str | None = None, password: str | None = None):
        # Avoid duplicate initialization of the same instance
        if hasattr(self, "_initialized") and self._initialized:
            return

        self.base_url = (
            base_url
            or os.getenv("RCABENCH_BASE_URL")
            or get_config(env_mode=os.environ.get("ENV_MODE", "dev")).base_url
        )
        self.username = username or os.getenv("RCABENCH_USERNAME")
        self.password = password or os.getenv("RCABENCH_PASSWORD")

        assert self.username is not None, "username or RCABENCH_USERNAME is not set"
        assert self.password is not None, "password or RCABENCH_PASSWORD is not set"
        assert self.base_url is not None, "base_url or RCABENCH_BASE_URL is not set"

        self.instance_key = (self.base_url, self.username)
        self._initialized = True

    def __enter__(self):
        # Check if there is already a valid session
        if self.instance_key not in self._sessions or not self._is_session_valid():
            self._login()
        return self._get_authenticated_client()

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Do not close session, maintain singleton state
        pass

    def _is_session_valid(self):
        """Check if the current session is valid"""
        session_data = self._sessions.get(self.instance_key)
        if not session_data:
            return False

        # More complex session validity checks can be added here, such as checking if token is expired
        # Currently simply check if access_token exists
        return session_data.get("access_token") is not None

    def _login(self):
        config = Configuration(host=self.base_url)
        with ApiClient(config) as api_client:
            auth_api = AuthenticationApi(api_client)
            assert self.username is not None
            assert self.password is not None
            login_request = DtoLoginRequest(username=self.username, password=self.password)
            response = auth_api.api_v2_auth_login_post(login_request)
            assert response.data is not None

            # Store session information in class-level cache
            self._sessions[self.instance_key] = {"access_token": response.data.token, "api_client": None}

    def _get_authenticated_client(self):
        session_data = self._sessions.get(self.instance_key)
        if not session_data:
            self._login()
            session_data = self._sessions[self.instance_key]

        # If api_client has not been created or needs to be updated, create a new one
        if not session_data.get("api_client"):
            auth_config = Configuration(
                host=self.base_url,
                api_key={"BearerAuth": session_data["access_token"]},
                api_key_prefix={"BearerAuth": "Bearer"},
            )
            session_data["api_client"] = ApiClient(auth_config)

        return session_data["api_client"]

    def get_client(self):
        if self.instance_key not in self._sessions or not self._is_session_valid():
            self._login()
        return self._get_authenticated_client()

    @classmethod
    def clear_sessions(cls):
        cls._sessions.clear()
        cls._instances.clear()


def get_evaluation_by_datapack(
    algorithm: str, datapack: str, tag: str | None = None, base_url: str | None = None
) -> list[DtoAlgorithmDatapackResp]:
    base_url = base_url or os.getenv("RCABENCH_BASE_URL")
    assert base_url is not None, "base_url or RCABENCH_BASE_URL is not set"
    assert tag, "Tag must be specified."

    with RCABenchClient(base_url=base_url) as client:
        api = EvaluationApi(client)
        resp = api.api_v2_evaluations_datapacks_post(
            request=DtoDatapackEvaluationBatchReq(
                items=[DtoAlgorithmDatapackReq(algorithm=algorithm, datapack=datapack, tag=tag)]
            )
        )

    assert resp.code is not None and resp.code < 300, f"Failed to get evaluation: {resp.message}"
    assert resp.data is not None
    return resp.data


def get_evaluation_by_dataset(
    algorithm: str,
    dataset: str,
    dataset_version: str | None = None,
    tag: str | None = None,
    base_url: str | None = None,
) -> list[DtoAlgorithmDatasetResp]:
    base_url = base_url or os.getenv("RCABENCH_BASE_URL")
    assert base_url is not None, "base_url or RCABENCH_BASE_URL is not set"

    with RCABenchClient(base_url=base_url) as client:
        api = EvaluationApi(client)
        resp = api.api_v2_evaluations_datasets_post(
            request=DtoDatasetEvaluationBatchReq(
                items=[
                    DtoAlgorithmDatasetReq(
                        algorithm=algorithm,
                        dataset=dataset,
                        dataset_version=dataset_version,
                        tag=tag,
                    )
                ]
            )
        )

    assert resp.code is not None and resp.code < 300, f"Failed to get evaluation: {resp.message}"
    assert resp.data is not None
    return resp.data


def get_datapacks_from_dataset_id(
    dataset_id: int | None = None,
) -> tuple[list[DtoInjectionV2Response], str, str]:
    with RCABenchClient() as client:
        assert dataset_id is not None
        injections: list[DtoInjectionV2Response] = []
        api = DatasetsApi(client)
        resp = api.api_v2_datasets_id_get(id=dataset_id, include_injections=True)
        if not resp or not resp.data or not resp.data.injections:
            raise ValueError(f"No injections found for dataset {dataset_id}")
        injections = resp.data.injections

        assert resp.data.name is not None and resp.data.version is not None
        return injections, resp.data.name, resp.data.version



"""
Anomaly detection configuration classes
Define independent configuration parameters
for each detection method
"""

from dataclasses import dataclass, field

from .types import DetectionMethod, MetricType


# ===== Base Configuration Class =====
@dataclass
class DetectionConfig:
    """Base class for detection configuration"""

    enabled: bool = True


# ===== Specific Detection Method Configurations =====
@dataclass
class ThresholdConfig(DetectionConfig):
    """
    Threshold detection configuration

    Suitable for metrics with explicit
    upper and lower limits
    """

    high_threshold: float | None = None
    low_threshold: float | None = None
    severity_critical_ratio: float = 2.0
    severity_high_ratio: float = 1.5
    severity_medium_ratio: float = 1.2


@dataclass
class StatisticalConfig(DetectionConfig):
    """
    Statistical detection configuration

    Statistical anomaly detection based on
    Z-score and IQR
    """

    z_score_threshold: float = 3.0
    iqr_multiplier: float = 1.5
    severity_zscore_critical: float = 4.0
    severity_zscore_high: float = 3.5
    severity_zscore_medium: float = 3.0
    severity_confidence_high: float = 0.8
    severity_confidence_medium: float = 0.5


@dataclass
class TrendConfig(DetectionConfig):
    """
    Trend detection configuration

    Detects spike and drop
    anomaly patterns
    """

    window_size: int = 10
    spike_threshold_multiplier: float = 2.0
    severity_spike_critical: float = 4.0
    severity_spike_high: float = 3.0
    severity_spike_medium: float = 2.0


@dataclass
class EnhancedLatencyConfig(DetectionConfig):
    """
    Enhanced latency detection configuration

    Advanced detection methods specifically
    for latency metrics
    """

    # Basic configuration
    percentile_type: str = "avg"  # Select percentile type for detection:
    # "avg"(average), "p90"(90th percentile),
    # "p95"(95th percentile), "p99"(99th percentile)
    hard_timeout_threshold: float = 15.0  # Hard timeout threshold(seconds),
    # exceeding this value directly determines as critical anomaly,
    # used to capture extreme timeout situations

    # Stability threshold configuration:
    # used to evaluate latency stability under different percentiles
    stability_thresholds: dict[str, float] = field(
        default_factory=lambda: {
            "avg": 2.0,  # Average latency stability threshold(seconds),
            # exceeding this value considers latency unstable
            "p90": 3.0,  # P90 latency stability threshold(seconds),
            # used to detect high percentile latency fluctuations
            "p95": 5.0,  # P95 latency stability threshold(seconds),
            # detect stricter latency requirements
            "p99": 6.0,  # P99 latency stability threshold(seconds),
            # detect extreme latency situations
        }
    )

    # === Baseline filtering rule configuration ===
    enable_baseline_filtering: bool = (
        True  # Whether to enable baseline filtering,
        # used to filter out anomaly detection with high baseline latency
    )
    baseline_avg_threshold: float = 1.0  # Baseline average latency threshold
    # (seconds), baseline data exceeding this value will be filtered
    # to avoid false positives
    baseline_p99_threshold: float = (
        5.0  # Baseline P99 latency threshold(seconds),
        # used to filter data with extremely high latency in baseline
    )

    # === Adaptive threshold configuration ===
    enable_adaptive_thresholds: bool = True  # Whether to enable adaptive
    # threshold adjustment, dynamically adjust detection sensitivity
    # based on baseline size
    small_baseline_threshold: float = 0.5  # Small baseline latency threshold
    # (seconds), when baseline latency is less than this value,
    # use absolute threshold detection
    absolute_anomaly_threshold: float = 2.0  # Absolute anomaly threshold
    # (seconds), used for absolute latency anomaly detection
    # in small baseline situations

    # === Base multiplier threshold configuration ===
    # Base multiplier thresholds for different percentiles,
    # used to calculate anomaly detection multipliers
    # relative to baseline
    base_multiplier_thresholds: dict[str, float] = field(
        default_factory=lambda: {
            "avg": 3.0,  # Average latency base multiplier:
            # triggers anomaly detection when current latency
            # exceeds 3x baseline
            "p90": 6.0,  # P90 latency base multiplier:
            # slightly higher multiplier considering natural
            # fluctuations of high percentiles
            "p95": 7.5,  # P95 latency base multiplier:
            # higher tolerance, reduces false positives for P95 latency
            "p99": 8.0,  # P99 latency base multiplier:
            # highest tolerance, P99 itself is extreme value statistics
        }
    )

    """
    # === Percentile baseline filtering thresholds ===
    # Skip anomaly detection when baseline latency exceeds these thresholds to avoid false positives
    # in high latency environments
    # Average latency baseline filtering threshold(seconds), skip detection when baseline exceeds 1 second
     # P90 latency baseline filtering threshold(seconds), skip detection when baseline exceeds 1.5 seconds
     # P95 latency baseline filtering threshold(seconds), skip detection when baseline exceeds 2 seconds
      # P99 latency baseline filtering threshold(seconds), skip detection when baseline exceeds 3 seconds
    """

    percentile_baseline_thresholds: dict[str, float] = field(
        default_factory=lambda: {
            "avg": 1.0,
            "p90": 1.5,
            "p95": 2.0,
            "p99": 3.0,
        }
    )

    """
    # === Percentile absolute anomaly thresholds ===
    # Use absolute thresholds for anomaly detection when baseline latency is small, avoid false positives 
    # from high multipliers of small values
    # Average latency absolute anomaly threshold(seconds), directly determine as anomaly when current 
    # latency exceeds 2 seconds
    # P90 latency absolute anomaly threshold(seconds), adapt to P90 statistical characteristics
    # P95 latency absolute anomaly threshold(seconds), higher absolute threshold
    # P99 latency absolute anomaly threshold(seconds), highest absolute threshold, avoid natural extremes of P99
    """
    percentile_absolute_thresholds: dict[str, float] = field(
        default_factory=lambda: {
            "avg": 2.0,
            "p90": 4.0,
            "p95": 4.5,
            "p99": 5.0,
        }
    )

    # === Percentile minimum multiplier limits ===
    # Ensure that even under adaptive adjustment, detection multipliers will not be lower than these minimum values,
    # guaranteeing the lower limit of detection sensitivity
    percentile_min_multipliers: dict[str, float] = field(
        default_factory=lambda: {
            "avg": 1.5,  # Average latency minimum detection multiplier, at least 1.5x to trigger anomaly
            "p90": 4,  # P90 latency minimum detection multiplier, considering P90 volatility
            "p95": 4.5,  # P95 latency minimum detection multiplier, more conservative detection strategy
            "p99": 5.0,  # P99 latency minimum detection multiplier, most conservative detection, avoid P99 fp
        }
    )


@dataclass
class SuccessRateConfig(DetectionConfig):
    """
    Success rate detection configuration

    Detects success rate drops based on
    statistical significance tests
    """

    min_normal_count: int = 10
    min_abnormal_count: int = 5
    min_rate_drop: float = 0.03
    significance_threshold: float = 0.05
    min_relative_drop: float = 0.1


# ===== Composite Configuration Class =====
@dataclass
class MetricDetectionConfig:
    """
    Metric detection configuration

    Combines multiple detection methods to
    serve specific metric types
    """

    metric_type: MetricType
    enabled_methods: dict[DetectionMethod, DetectionConfig] = field(default_factory=dict)

    def get_method_config(self, method: DetectionMethod) -> DetectionConfig | None:
        """Get the configuration for the specified method"""
        return self.enabled_methods.get(method)

    def is_method_enabled(self, method: DetectionMethod) -> bool:
        """Check if the method is enabled"""
        config = self.enabled_methods.get(method)
        return config is not None and config.enabled

    def add_method(self, method: DetectionMethod, config: DetectionConfig) -> None:
        """Add a detection method"""
        self.enabled_methods[method] = config

    def remove_method(self, method: DetectionMethod) -> None:
        """Remove a detection method"""
        self.enabled_methods.pop(method, None)

    def enable_method(self, method: DetectionMethod) -> bool:
        """Enable a detection method"""
        config = self.enabled_methods.get(method)
        if config is not None:
            config.enabled = True
            return True
        return False

    def disable_method(self, method: DetectionMethod) -> bool:
        """Disable a detection method"""
        config = self.enabled_methods.get(method)
        if config is not None:
            config.enabled = False
            return True
        return False



from ....vendor.RCAEval.baro import baro
from ...algorithms.spec import Algorithm, AlgorithmAnswer, AlgorithmArgs
from ._common import SimpleMetricsAdapter


class Baro(Algorithm):
    def needs_cpu_count(self) -> int | None:
        return 4

    def __call__(self, args: AlgorithmArgs) -> list[AlgorithmAnswer]:
        adapter = SimpleMetricsAdapter(baro)
        return adapter(args)



import os

import neo4j


def get_neo4j_driver(
    uri: str | None = None,
    username: str | None = None,
    password: str | None = None,
) -> neo4j.Driver:
    """
    Establish a connection to Neo4j database

    Args:
        uri: Neo4j database URI, defaults to environment variable NEO4J_URI or "bolt://localhost:7687"
        username: Neo4j username, defaults to environment variable NEO4J_USERNAME or "neo4j"
        password: Neo4j password, defaults to environment variable NEO4J_PASSWORD

    Returns:
        Neo4j driver instance
    """
    # Use parameters or fallback to environment variables or defaults
    uri = uri or os.environ.get("NEO4J_URI", "bolt://localhost:7687")
    username = username or os.environ.get("NEO4J_USERNAME", "neo4j")
    password = password or os.environ.get("NEO4J_PASSWORD", "your_password")

    driver = neo4j.GraphDatabase.driver(uri, auth=(username, password))
    return driver



"""Specification for sampler experiments."""

from pathlib import Path

from ...datasets.spec import get_datapack_folder
from ..spec import SamplingMode


def get_sampler_output_folder(
    dataset: str, datapack: str, sampler: str, sampling_rate: float, mode: SamplingMode
) -> Path:
    """
    Get the output folder for sampler results.

    Format: {input_folder}/sampled/{sampler}_{sampling_rate}_{mode}
    Creates a sampled subdirectory in the datapack's input folder.
    """
    input_folder = get_datapack_folder(dataset, datapack)
    mode_str = mode.value
    rate_str = f"{sampling_rate:.3f}".rstrip("0").rstrip(".")
    folder_name = f"{sampler}_{rate_str}_{mode_str}"
    return input_folder / "sampled" / folder_name



import traceback

import typer

from ..clients.clickhouse import get_clickhouse_client
from ..clients.rcabench_ import get_rcabench_openapi_client
from ..config import get_config
from ..logging import logger, timeit

app = typer.Typer()


@app.command()
@timeit()
def ping_clickhouse() -> None:
    with get_clickhouse_client() as client:
        assert client.ping(), "clickhouse should be reachable"
        logger.info("clickhouse is reachable")


@app.command()
@timeit()
def ping_rcabench() -> None:
    from rcabench.openapi import AlgorithmApi

    api = AlgorithmApi(get_rcabench_openapi_client())
    resp = api.api_v1_algorithms_get()
    assert resp.data is not None

    logger.info("rcabench is reachable")


@app.command()
@timeit()
def test() -> None:
    logger.info("Testing rcabench-platform environment...")

    try:
        ping_clickhouse()
    except Exception as e:
        traceback.print_exc()
        logger.error(f"ClickHouse ping failed: {e}")

    try:
        ping_rcabench()
    except Exception as e:
        traceback.print_exc()
        logger.error(f"RCABench ping failed: {e}")

    try:
        config = get_config()
        config.data.stat()
        logger.opt(colors=True).info(f"config.data is found: <green>{config.data}</green>")
    except Exception as e:
        traceback.print_exc()
        logger.error(f"config.data is not found: {e}")

    logger.info("Hello from rcabench-platform!")



from dataclasses import asdict, dataclass, field
from typing import Any

import polars as pl
from rcabench.openapi import HandlerNode

from ..logging import logger
from .data_prepare import Item, get_conf

# Constants
MIN_DEPTH_FOR_RANGE = 2
DATAPACK_METRIC_PREFIX = "datapack_metric_"

# Column names used throughout the module
FAULT_TYPE_COL = "fault_type"
INJECTED_SERVICE_COL = "injected_service"
IS_PAIR_COL = "is_pair"


@dataclass
class PairStats:
    in_degree: int = 0
    out_degree: int = 0


@dataclass
class Distribution:
    faults: dict[str, int] = field(default_factory=dict)
    services: dict[str, int] = field(default_factory=dict)
    pairs: dict[str, PairStats] = field(default_factory=dict)

    fault_services: dict[str, dict[str, int]] = field(default_factory=dict)
    fault_service_attribute_coverages: dict[str, dict[str, float]] = field(default_factory=dict)
    fault_service_metrics: dict[str, dict[str, dict[str, dict[str, int]]]] = field(default_factory=dict)

    fault_pairs: dict[str, dict[str, int]] = field(default_factory=dict)
    fault_pair_attribute_coverages: dict[str, dict[str, float]] = field(default_factory=dict)
    fault_pair_metrics: dict[str, dict[str, dict[str, dict[str, int]]]] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        return {
            "faults": self.faults,
            "services": self.services,
            "pairs": {k: asdict(v) for k, v in self.pairs.items()},
            "fault_services": self.fault_services,
            "fault_service_attribute_coverages": self.fault_service_attribute_coverages,
            "fault_service_metrics": {
                k: {sk: sv for sk, sv in v.items()} for k, v in self.fault_service_metrics.items()
            },
            "fault_pairs": self.fault_pairs,
            "fault_pair_attribute_coverages": self.fault_pair_attribute_coverages,
            "fault_pair_metrics": {k: {sk: sv for sk, sv in v.items()} for k, v in self.fault_pair_metrics.items()},
        }


def get_datapacks_distribution(df: pl.DataFrame, metrics: list[str]) -> Distribution:
    if df.height == 0:
        return Distribution()

    distribution = Distribution()

    # Basic distributions using DataFrame operations
    distribution.faults = _get_single_column_distribution(df, FAULT_TYPE_COL)
    distribution.services = _get_single_column_distribution(df, INJECTED_SERVICE_COL)
    distribution.pairs = get_pairs_distribution(df)

    # Composite distributions
    distribution.fault_services = _get_fault_target_distribution(df, is_pair=False)
    distribution.fault_pairs = _get_fault_target_distribution(df, is_pair=True)

    # Datapack metric distributions
    distribution.fault_service_metrics = get_fault_service_metrics_distribution(df, metrics)

    return distribution


def _validate_dataframe_columns(df: pl.DataFrame, required_columns: list[str]) -> bool:
    return df.height > 0 and all(col in df.columns for col in required_columns)


def _get_single_column_distribution(df: pl.DataFrame, column_name: str) -> dict[str, int]:
    if not _validate_dataframe_columns(df, [column_name]):
        return {}

    counts = df.group_by(column_name).agg(pl.len().alias("count"))
    return {row[column_name]: row["count"] for row in counts.iter_rows(named=True)}


def get_pairs_distribution(df: pl.DataFrame) -> dict[str, PairStats]:
    if not _validate_dataframe_columns(df, [INJECTED_SERVICE_COL, IS_PAIR_COL]):
        return {}

    # Filter for pairs only
    pairs_df = df.filter(pl.col(IS_PAIR_COL))

    if pairs_df.height == 0:
        return {}

    # Split pairs and calculate degrees
    pairs_with_split = pairs_df.with_columns(
        [
            pl.col(INJECTED_SERVICE_COL).str.split("->").list.get(0).alias("source_service"),
            pl.col(INJECTED_SERVICE_COL).str.split("->").list.get(1).alias("target_service"),
        ]
    ).filter((pl.col("source_service").is_not_null()) & (pl.col("target_service").is_not_null()))

    if pairs_with_split.height == 0:
        return {}

    # Calculate out degrees (source services)
    out_degrees = (
        pairs_with_split.group_by("source_service")
        .agg(pl.len().alias("out_degree"))
        .rename({"source_service": "service"})
    )

    # Calculate in degrees (target services)
    in_degrees = (
        pairs_with_split.group_by("target_service")
        .agg(pl.len().alias("in_degree"))
        .rename({"target_service": "service"})
    )

    # Merge degrees
    all_services = out_degrees.join(in_degrees, on="service", how="full").fill_null(0)

    pairs_stats = {}
    for row in all_services.iter_rows(named=True):
        pairs_stats[row["service"]] = PairStats(in_degree=row.get("in_degree", 0), out_degree=row.get("out_degree", 0))

    return pairs_stats


def _get_fault_target_distribution(df: pl.DataFrame, is_pair: bool) -> dict[str, dict[str, int]]:
    if not _validate_dataframe_columns(df, [FAULT_TYPE_COL, INJECTED_SERVICE_COL]):
        return {}

    # Filter based on is_pair flag
    filtered_df = df.filter(pl.col(IS_PAIR_COL) == is_pair)

    if filtered_df.height == 0:
        return {}

    fault_target_counts = filtered_df.group_by([FAULT_TYPE_COL, INJECTED_SERVICE_COL]).agg(pl.len().alias("count"))

    result = {}
    for row in fault_target_counts.iter_rows(named=True):
        fault_type = row[FAULT_TYPE_COL]
        target = row[INJECTED_SERVICE_COL]
        count = row["count"]

        if fault_type not in result:
            result[fault_type] = {}
        result[fault_type][target] = count

    return result


class _NodeProcessor:
    @staticmethod
    def __call__(node: HandlerNode, **kwargs) -> Any:
        raise NotImplementedError

    @staticmethod
    def combine(results: list) -> Any:
        raise NotImplementedError


class _RangeProcessor(_NodeProcessor):
    @staticmethod
    def __call__(node: HandlerNode, **kwargs) -> int:
        return node.range[1] - node.range[0] + 1 if node.range else 0

    @staticmethod
    def combine(results: list[int]) -> int:
        return sum(results)


class _CoverageProcessor(_NodeProcessor):
    @staticmethod
    def __call__(node: HandlerNode, **kwargs) -> dict[str, bool]:
        return {f"{kwargs['key']}-{node.value}": True}

    @staticmethod
    def combine(results: list[dict]) -> dict[str, bool]:
        combined = {}
        for result in results:
            combined.update(result)
        return combined


def _traverse_node(node: HandlerNode, key: str, processor: _NodeProcessor) -> Any:
    int_key = int(key)

    if node.children is None:
        if int_key > MIN_DEPTH_FOR_RANGE:
            return processor(node, key=key)
        else:
            # Return appropriate default value based on processor type
            if isinstance(processor, _RangeProcessor):
                return 0
            elif isinstance(processor, _CoverageProcessor):
                return {}
            else:
                return {}

    results = []
    for child_key, child_node in node.children.items():
        results.append(_traverse_node(child_node, child_key, processor))

    return processor.combine(results)


def _get_fault_target_coverages(coverage_df: pl.DataFrame, is_pair: bool) -> dict[str, dict[str, float]]:
    if coverage_df.height == 0:
        return {}

    # Filter based on is_pair flag
    filtered_coverage = coverage_df.filter(pl.col(IS_PAIR_COL) == is_pair)

    result = {}
    for row in filtered_coverage.iter_rows(named=True):
        fault_type = row[FAULT_TYPE_COL]
        target = row[INJECTED_SERVICE_COL]
        coverage = row["coverage"]

        if fault_type not in result:
            result[fault_type] = {}
        result[fault_type][target] = coverage

    return result


def get_fault_service_coverages(coverage_df: pl.DataFrame) -> dict[str, dict[str, float]]:
    return _get_fault_target_coverages(coverage_df, is_pair=False)


def get_fault_pair_coverages(coverage_df: pl.DataFrame) -> dict[str, dict[str, float]]:
    return _get_fault_target_coverages(coverage_df, is_pair=True)


def get_fault_service_metrics_distribution(
    df: pl.DataFrame, metrics: list[str]
) -> dict[str, dict[str, dict[str, dict[str, int]]]]:
    if df.height == 0:
        return {}

    # Filter for pairs only (as per original logic)
    pairs_df = df.filter(pl.col(IS_PAIR_COL))

    if pairs_df.height == 0:
        return {}

    # Find datapack metric columns
    datapack_metric_cols = [col for col in df.columns if col.startswith(DATAPACK_METRIC_PREFIX)]

    # Filter metrics that actually exist in the DataFrame
    available_metrics = []
    for metric in metrics:
        metric_col = f"{DATAPACK_METRIC_PREFIX}{metric}"
        if metric_col in datapack_metric_cols:
            available_metrics.append(metric)

    if not available_metrics:
        return {}

    result = {}

    # Group by fault type and service to get metric distributions
    for (fault_type_obj, service_obj), group_df in pairs_df.group_by([FAULT_TYPE_COL, INJECTED_SERVICE_COL]):
        fault_type = str(fault_type_obj)
        service = str(service_obj)

        if fault_type not in result:
            result[fault_type] = {}

        if service not in result[fault_type]:
            result[fault_type][service] = {}

        # For each metric, calculate value distribution
        for metric in available_metrics:
            metric_col = f"{DATAPACK_METRIC_PREFIX}{metric}"

            # Get value counts for this metric
            metric_values = group_df.select(metric_col).to_series()
            value_counts = metric_values.value_counts().sort("count", descending=True)

            # Convert to the expected format
            distribution_dict = {}
            for row in value_counts.iter_rows(named=True):
                value = str(row[metric_col])
                count = row["count"]
                distribution_dict[value] = count

            result[fault_type][service][metric] = distribution_dict

    return result



from dataclasses import asdict
from typing import Any

import polars as pl

from ...logging import logger
from .defintion import SDG, DepEdge, PlaceKind, PlaceNode


def dump_place_indicators(sdg: SDG) -> pl.DataFrame:
    rows = []
    data_keys: dict[str, Any] = {}

    for node in sdg.iter_nodes():
        row_base = {
            "node.id": node.id,
            "place.kind": node.kind,
            "place.name": node.self_name,
        }

        if len(node.indicators) == 0:
            row = {
                "indicator.name": None,
                "indicator.df.length": None,
                **row_base,
            }
            continue

        for indicator in node.indicators.values():
            row = {
                "indicator.name": indicator.name,
                "indicator.df.length": len(indicator.df),
                **row_base,
            }

            for k, v in indicator.data.items():
                if v is None:
                    continue

                assert isinstance(v, (int, float, str))
                row[f"indicator.data.{k}"] = v

                prev_type = data_keys.get(k)
                if prev_type is None or prev_type is type(None):
                    data_keys[k] = type(v)
                else:
                    assert isinstance(v, prev_type), f"Type mismatch for {k}: {v} ({type(v)}) vs {prev_type}"

            rows.append(row)

    schema = {
        "node.id": pl.UInt32,
        "place.kind": pl.String,
        "place.name": pl.String,
        "indicator.name": pl.String,
        "indicator.df.length": pl.UInt64,
    }

    for k, v in data_keys.items():
        if v is str:
            dt = pl.Utf8
        elif v is int:
            dt = pl.Int64
        elif v is float:
            dt = pl.Float64
        else:
            logger.error(f"Unknown type for {k}: {v}")
            raise NotImplementedError
        schema[f"indicator.data.{k}"] = dt

    df = pl.DataFrame(rows, schema=schema)
    return df


def replace_large_set(data: dict[str, Any]):
    for k, v in data.items():
        if isinstance(v, set) and len(v) >= 16:
            data[k] = {"len": len(v), "content": "<large set>"}
        elif isinstance(v, dict):
            replace_large_set(v)


def replace_large_df(data: dict[str, Any]):
    for k, v in data.items():
        if isinstance(v, pl.DataFrame):
            data[k] = {"len": len(v), "schema": v.schema, "content": "<large DataFrame>"}
        elif isinstance(v, dict):
            replace_large_df(v)


def dump_edge_data(edge: DepEdge) -> dict[str, Any]:
    data = asdict(edge)
    replace_large_set(data)
    replace_large_df(data)
    return data


def dump_node_data(node: PlaceNode) -> dict[str, Any]:
    data = asdict(node)
    replace_large_set(data)
    replace_large_df(data)
    return data


def dump_span_tree(sdg: SDG, top: PlaceNode, trace_id) -> dict[str, Any] | None:
    assert top.kind == PlaceKind.function
    indicator = top.indicators["duration"]

    trace = indicator.df.filter(pl.col("trace_id") == trace_id)
    if len(trace) == 0:
        return

    self_duration_sum = trace["value"].sum()
    self_duration_sum /= 1e6  # ns -> ms

    children = []

    children_duration_sum = 0

    for edge in sdg.out_edges(top.id):
        dst = sdg.get_node_by_id(edge.dst_id)
        if dst.kind == PlaceKind.function:
            ans = dump_span_tree(sdg, dst, trace_id)
            if ans:
                children.append(ans)
                children_duration_sum += ans["self.duration.sum"]

    min_time = trace["time"].min()
    max_time = trace["time"].max()

    return {
        "name": top.self_name,
        "min_start_time": min_time,
        "max_start_time": max_time,
        "self.duration.sum": self_duration_sum,
        "children.duration.sum": children_duration_sum,
        "children": children,
    }



import os
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path


@dataclass(kw_only=True)
class Config:
    env_mode: str

    data: Path
    output: Path
    temp: Path

    base_url: str


_DEFAULT_DATA_ROOT = Path(os.getenv("DATA_ROOT", "data/rcabench-platform-v2"))
_DEFAULT_OUTPUT_ROOT = Path(os.getenv("OUTPUT_ROOT", "output/rcabench-platform-v2"))
_DEFAULT_TEMP_ROOT = Path(os.getenv("TEMP_ROOT", "temp"))

_DEBUG_CONFIG = Config(
    env_mode="debug",
    data=_DEFAULT_DATA_ROOT,
    output=_DEFAULT_OUTPUT_ROOT,
    temp=_DEFAULT_TEMP_ROOT,
    base_url="http://127.0.0.1:8082",
)

_DEV_CONFIG = Config(
    env_mode="dev",
    data=_DEFAULT_DATA_ROOT,
    output=_DEFAULT_OUTPUT_ROOT,
    temp=_DEFAULT_TEMP_ROOT,
    base_url="http://10.10.10.161:8082",
)

_PROD_CONFIG = Config(
    env_mode="prod",
    data=_DEFAULT_DATA_ROOT,
    output=_DEFAULT_OUTPUT_ROOT,
    temp=_DEFAULT_TEMP_ROOT,
    base_url="http://10.10.10.220:32080",
)

_CONFIG_CLASSES = {
    "debug": _DEBUG_CONFIG,
    "dev": _DEV_CONFIG,
    "prod": _PROD_CONFIG,
}

_CONFIG = None


def get_config(env_mode: str | None = None) -> Config:
    if env_mode is not None:
        return _CONFIG_CLASSES[env_mode]

    global _CONFIG

    if _CONFIG is None:
        env_mode = os.getenv("ENV_MODE", "prod").lower()
        _CONFIG = _CONFIG_CLASSES[env_mode]

    return _CONFIG


def set_config(config: Config):
    global _CONFIG
    _CONFIG = config


@contextmanager
def current_config(config: Config):
    global _CONFIG

    old_config = _CONFIG
    _CONFIG = config

    try:
        yield
    finally:
        _CONFIG = old_config



import multiprocessing
import multiprocessing.pool
import os
import time
import traceback
from collections.abc import Callable, Sequence
from typing import Any, Literal, TypeVar

from tqdm.auto import tqdm

from ..algorithms.spec import global_algorithm_registry, set_global_algorithm_registry
from ..config import get_config, set_config
from ..logging import get_real_logger, logger, set_real_logger, timeit
from ..samplers.spec import global_sampler_registry, set_global_sampler_registry


def set_cpu_limit_outer(n: int | None) -> None:
    if n is None:
        return

    os.environ["POLARS_MAX_THREADS"] = str(n)


def set_cpu_limit_inner(n: int | None) -> None:
    if n is None:
        return

    try:
        import torch  # type: ignore

        torch.set_num_threads(n)
    except ImportError:
        pass


def initializers(*, cpu_limit: int | None = None) -> list[tuple[Callable, Any]]:
    ans = [
        (set_cpu_limit_inner, (cpu_limit,)),
        # (set_real_logger, (get_real_logger(),)),
        (set_config, (get_config(),)),
        (set_global_algorithm_registry, (global_algorithm_registry(),)),
        (set_global_sampler_registry, (global_sampler_registry(),)),
    ]

    return ans


def call_initializers(init_list: list[tuple[Callable, Any]]) -> None:
    for func, args in init_list:
        func(*args)


R = TypeVar("R")


def _fmap(
    mode: Literal["threadpool", "processpool"],
    tasks: Sequence[Callable[[], R]],
    *,
    parallel: int,
    ignore_exceptions: bool,
    cpu_limit_each: int | None,
) -> list[R]:
    if cpu_limit_each is not None:
        assert mode == "processpool", "cpu_limit is only supported for processpool mode"
        assert cpu_limit_each > 0, "cpu_limit must be greater than 0"

    if not isinstance(tasks, list):
        tasks = list(tasks)

    if len(tasks) == 0:
        return []

    if parallel is None or parallel > 1:
        num_workers = parallel or multiprocessing.cpu_count()
        num_workers = min(num_workers, len(tasks))
    else:
        num_workers = 1

    logger_ = logger.opt(depth=2)

    if num_workers > 1:
        if mode == "threadpool":
            pool = multiprocessing.pool.ThreadPool(
                processes=num_workers,
            )
        elif mode == "processpool":
            set_cpu_limit_outer(cpu_limit_each)
            pool = multiprocessing.get_context("spawn").Pool(
                processes=num_workers,
                initializer=call_initializers,
                initargs=(initializers(),),
            )
        else:
            raise ValueError(f"Unknown mode: {mode}")

        with pool:
            asyncs = [pool.apply_async(task) for task in tasks]
            finished = [False] * len(asyncs)
            index_results: list[tuple[int, R]] = []
            exception_count = 0

            with tqdm(total=len(asyncs), desc=f"fmap_{mode}") as pbar:
                while not all(finished):
                    for i, async_ in enumerate(asyncs):
                        if finished[i]:
                            continue
                        if not async_.ready():
                            continue
                        try:
                            result = async_.get(timeout=0.1)
                            finished[i] = True
                            index_results.append((i, result))
                            pbar.update(1)
                        except multiprocessing.TimeoutError:
                            continue
                        except Exception as e:
                            exception_count += 1
                            finished[i] = True
                            pbar.update(1)
                            if ignore_exceptions:
                                traceback.print_exc()
                                logger_.error("Exception in task {}: {}", i, e)
                            else:
                                raise e
                    pbar.update(0)
                    time.sleep(1)

        index_results.sort(key=lambda x: x[0])
        results = [result for _, result in index_results]
    else:
        results = []
        exception_count = 0
        for i, task in enumerate(tqdm(tasks, desc="fmap")):
            try:
                result = task()
                results.append(result)
            except Exception as e:
                exception_count += 1
                if ignore_exceptions:
                    traceback.print_exc()
                    logger_.error("Exception in task {}: {}", i, e)
                else:
                    raise e

    if exception_count > 0:
        logger_.warning(f"fmap_{mode} completed with {exception_count} exceptions.")

    logger_.debug(f"fmap_{mode} completed with {len(results)} results in {len(tasks)} tasks.")

    return results


@timeit(log_args=False)
def fmap_threadpool(
    tasks: Sequence[Callable[[], R]],
    *,
    parallel: int,
    ignore_exceptions: bool = False,
) -> list[R]:
    return _fmap(
        "threadpool",
        tasks,
        parallel=parallel,
        ignore_exceptions=ignore_exceptions,
        cpu_limit_each=None,
    )


@timeit(log_args=False)
def fmap_processpool(
    tasks: Sequence[Callable[[], R]],
    *,
    parallel: int,
    ignore_exceptions: bool = False,
    cpu_limit_each: int | None = None,
) -> list[R]:
    return _fmap(
        "processpool",
        tasks,
        parallel=parallel,
        ignore_exceptions=ignore_exceptions,
        cpu_limit_each=cpu_limit_each,
    )



from pathlib import Path
from typing import Any

import polars as pl

from .convert import DatapackLoader, DatasetLoader, Label


def convert_traces(src: Path):
    assert src.exists(), src

    lf = pl.scan_csv(src, infer_schema_length=50000)

    lf = lf.select(
        pl.from_epoch("startTime", time_unit="us").dt.replace_time_zone("UTC").alias("time"),
        pl.col("traceID").alias("trace_id"),
        pl.col("spanID").alias("span_id"),
        pl.col("serviceName").alias("service_name"),
        pl.col("operationName").alias("span_name"),
        pl.col("parentSpanID").alias("parent_span_id"),
        pl.col("duration").cast(pl.UInt64).mul(1000).alias("duration"),
        pl.col("statusCode").cast(pl.Int16, strict=False).alias("attr.status_code"),
    )

    return lf


def convert_metrics(src: Path):
    assert src.exists(), src

    lf = pl.scan_csv(src, infer_schema_length=50000)

    lf = lf.with_columns(pl.from_epoch("time", time_unit="s").dt.replace_time_zone("UTC").alias("time"))

    lf = lf.unpivot(
        on=None,
        index="time",
        variable_name="metric",
        value_name="value",
    )

    lf = lf.with_columns(
        pl.col("metric").str.split("_").alias("_split"),
    )

    lf = lf.with_columns(
        pl.col("_split").list.get(0).alias("service_name"),
        pl.col("_split").list.get(1).alias("metric"),
    )

    lf = lf.drop("_split")

    return lf


def convert_logs(src: Path):
    assert src.exists(), src

    lf = pl.scan_csv(src, infer_schema_length=50000)

    # Get column names to check for optional columns
    columns = lf.collect_schema().names()

    # Error keywords to search for in messages
    ERROR_KWS = ["error", "fail", "exception", "timeout", "refused"]

    # Base columns that should always exist
    base_select = [
        pl.from_epoch("timestamp", time_unit="ns").dt.replace_time_zone("UTC").alias("time"),
        pl.col("container_name").cast(pl.String).alias("service_name"),
        # Add empty trace_id and span_id columns
        pl.lit(None).cast(pl.String).alias("trace_id"),
        pl.lit(None).cast(pl.String).alias("span_id"),
    ]

    # Handle message column with optional error concatenation
    if "error" in columns:
        message_expr = (
            pl.when(pl.col("error").is_not_null())
            .then(pl.concat_str([pl.col("message"), pl.lit(" "), pl.col("error")]))
            .otherwise(pl.col("message"))
            .alias("message")
        )
    else:
        message_expr = pl.col("message")

    base_select.append(message_expr)

    # Handle level column - either use existing or parse from message
    if "level" in columns:
        base_select.append(pl.col("level").cast(pl.String))
    else:
        # Parse level from message - look for common log levels anywhere in the message
        # Keep null when extraction fails
        level_expr = pl.col("message").str.extract(r"\b(TRACE|DEBUG|INFO|WARN|WARNING|ERROR|FATAL)\b").alias("level")
        base_select.append(level_expr)

    # Add req_path with attr prefix if it exists
    if "req_path" in columns:
        base_select.append(pl.col("req_path").cast(pl.String).alias("attr.req_path"))

    # Add optional columns if they exist
    if "cluster_id" in columns:
        base_select.append(pl.col("cluster_id").cast(pl.UInt8).alias("attr.cluster_id"))

    if "log_template" in columns:
        base_select.append(pl.col("log_template").cast(pl.String).alias("attr.log_template"))

    # First apply the select to get the transformed columns
    lf = lf.select(base_select)

    # Then add attr.has_error based on the transformed message column
    # This ensures we're checking the complete message including any concatenated error info
    lf = lf.with_columns(
        pl.col("message").str.contains(f"(?i){'|'.join(ERROR_KWS)}").fill_null(False).alias("attr.has_error")
    )

    return lf


class RcaevalDatapackLoader(DatapackLoader):
    def __init__(self, src_folder: Path, dataset: str, datapack: str, service: str) -> None:
        self._src_folder = src_folder
        self._dataset = dataset
        self._datapack = datapack
        self._service = service

    def name(self) -> str:
        return self._datapack

    def labels(self) -> list[Label]:
        return [Label(level="service", name=self._service)]

    def data(self) -> dict[str, Any]:
        data = {
            "inject_time.txt": self._src_folder / "inject_time.txt",
            "simple_metrics.parquet": convert_metrics(self._src_folder / "simple_metrics.csv"),
            "logs.parquet": convert_logs(self._src_folder / "logs.csv"),
        }
        if (self._src_folder / "traces.csv").exists():
            data["traces.parquet"] = convert_traces(self._src_folder / "traces.csv")

        return data


class RcaevalDatasetLoader(DatasetLoader):
    def __init__(self, src_folder: Path, dataset: str):
        self._src_folder = src_folder
        self._dataset = dataset

        datapack_loaders = []

        for service_path in src_folder.iterdir():
            if not service_path.is_dir():
                continue

            for num_path in service_path.iterdir():
                if not num_path.is_dir():
                    continue

                service = service_path.name
                num = num_path.name
                datapack = f"{service}_{num}"

                if num == "multi-source-data":
                    continue

                loader = RcaevalDatapackLoader(
                    src_folder=num_path,
                    dataset=dataset,
                    datapack=datapack,
                    service=service.split("_")[0],
                )

                datapack_loaders.append(loader)

        self._datapack_loaders = datapack_loaders

    def name(self) -> str:
        return self._dataset

    def __len__(self) -> int:
        return len(self._datapack_loaders)

    def __getitem__(self, index: int) -> DatapackLoader:
        return self._datapack_loaders[index]



import subprocess
import tempfile
from abc import ABC, abstractmethod
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path


class Storage(ABC):
    @abstractmethod
    def upload_meta(self, dataset: str) -> None: ...

    @abstractmethod
    def upload_datapack(self, dataset: str, datapack: str) -> None: ...

    @abstractmethod
    def upload_dataset(self, dataset: str) -> None: ...

    @abstractmethod
    def download_meta(self, dataset: str) -> None: ...

    @abstractmethod
    def download_datapack(self, dataset: str, datapack: str) -> None: ...

    @abstractmethod
    def download_dataset(self, dataset: str) -> None: ...


@contextmanager
def zipped_folder(folder_path: Path, zip_name: str) -> Generator[tuple[Path, Path], None, None]:
    assert folder_path.is_dir()
    assert folder_path.parent.is_dir()

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        zip_path = tmp_path / f"{zip_name}.zip"

        subprocess.run(
            ["zip", "-r", str(zip_path), str(folder_path.name)],
            cwd=folder_path.parent,
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        sha256_path = zip_path.with_suffix(".sha256")
        subprocess.run(
            ["sha256sum", str(zip_path.name)],
            cwd=tmp_path,
            check=True,
            stdout=sha256_path.open("w"),
            stderr=subprocess.DEVNULL,
        )

        yield zip_path, sha256_path


@contextmanager
def unzipped_folder(zip_path: Path) -> Generator[Path, None, None]:
    assert zip_path.is_file()

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        subprocess.run(
            ["unzip", str(zip_path)],
            cwd=str(tmpdir),
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        folder = tmpdir / zip_path.stem
        assert folder.is_dir()

        yield folder



import datetime
import functools
import math
import traceback

import numpy as np
import polars as pl
import scipy.stats

from ...logging import logger, timeit
from .defintion import SDG, DepEdge, DepKind, Indicator, PlaceKind, PlaceNode


@timeit(log_args=False)
def calc_statistics(sdg: SDG):
    for node in sdg.iter_nodes():
        if node.kind == PlaceKind.function:
            calc_stat_for_function_node(node)
        elif node.kind == PlaceKind.pod:
            calc_stat_for_pod_node(node)
        elif node.kind == PlaceKind.container:
            calc_stat_for_container_node(node)

    for node in sdg.iter_nodes():
        if node.kind == PlaceKind.service:
            calc_stat_for_service_node(sdg, node)

    for edge in sdg.iter_edges():
        if edge.kind == DepKind.calls:
            calc_stat_for_calls_edge(sdg, edge)

    node_stat_names = set()
    for node in sdg.iter_nodes():
        collect_node_stat_names(node, node_stat_names)
    sdg.data["node_stat_names"] = sorted(node_stat_names)


STAT_PREFIX = ["stat.normal", "stat.anomal"]


def collect_node_stat_names(node: PlaceNode, names: set[str]) -> None:
    prefix = STAT_PREFIX[0]
    for k in node.data.keys():
        if k.startswith(prefix):
            stat_name = k[len(prefix) + 1 :]  # +1 for the dot
            names.add(stat_name)


def calc_stat_for_service_node(sdg: SDG, node: PlaceNode):
    assert node.kind == PlaceKind.service

    indicator_names = [
        "cpu",
        "mem",
        "diskio",
        "socket",
        "workload",
        "error",
        "latency-50",
        "latency-90",
    ]

    for name in indicator_names:
        indicator = node.indicators.get(name)
        if indicator is None:
            continue

        assert indicator.df.schema["value"] == pl.Float64
        assert indicator.df["value"].is_not_null().all()

        lf = indicator.df.lazy()
        normal_lf = lf.filter(pl.col("anomal") == 0)
        anomal_lf = lf.filter(pl.col("anomal") == 1)

        df_list = pl.collect_all([normal_lf, anomal_lf])

        value_list = [df["value"].to_numpy() for df in df_list]

        calc_regular_stat(node, indicator, value_list)

        node_stat = {
            "cpu": "cpu_usage",
            "mem": "memory_usage",
            "diskio": "diskio",
            "socket": "socket",
            "workload": "workload",
            "error": "error_rate",
            "latency-50": "latency_p50",
            "latency-90": "latency_p90",
        }

        for indicator_name, stat_name in node_stat.items():
            if indicator.name == indicator_name:
                mean = [indicator.data[f"{STAT_PREFIX[i]}.mean"] for i in range(2)]
                node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = mean[0]
                node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = mean[1]

    functions: list[PlaceNode] = []
    for edge in sdg.out_edges(node.id):
        if edge.kind == DepKind.includes:
            dst = sdg.get_node_by_id(edge.dst_id)
            assert dst.kind == PlaceKind.function
            functions.append(dst)

    if functions:
        error_rates: list[list[float]] = [[], []]
        for function in functions:
            values: list[float | None] = [function.data.get(f"{STAT_PREFIX[i]}.error_rate") for i in range(2)]
            if values[0] is None or values[1] is None:
                continue
            error_rates[0].append(values[0])
            error_rates[1].append(values[1])

        if error_rates[0] and error_rates[1]:
            node.data[f"{STAT_PREFIX[0]}.function_error_rate.mean"] = np.mean(error_rates[0])
            node.data[f"{STAT_PREFIX[1]}.function_error_rate.mean"] = np.mean(error_rates[1])


def calc_stat_for_pod_node(node: PlaceNode):
    assert node.kind == PlaceKind.pod

    indicator_names = [
        "k8s.pod.cpu.usage",
        "k8s.pod.memory.usage",
        "jvm.gc.duration:PS Scavenge:hist.sum",
    ]

    for name in indicator_names:
        indicator = node.indicators.get(name)
        if indicator is None:
            continue

        assert indicator.df.schema["value"] == pl.Float64
        assert indicator.df["value"].is_not_null().all()

        lf = indicator.df.lazy()
        normal_lf = lf.filter(pl.col("anomal") == 0)
        anomal_lf = lf.filter(pl.col("anomal") == 1)

        df_list = pl.collect_all([normal_lf, anomal_lf])

        value_list = [df["value"].to_numpy() for df in df_list]

        calc_regular_stat(node, indicator, value_list)

        if indicator.name == "k8s.pod.cpu.usage":
            stat_name = "cpu_usage"
            mean = [indicator.data[f"{STAT_PREFIX[i]}.mean"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = mean[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = mean[1]

        elif indicator.name == "k8s.pod.memory.usage":
            stat_name = "memory_usage"
            mean = [indicator.data[f"{STAT_PREFIX[i]}.mean"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = mean[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = mean[1]

        elif indicator.name == "jvm.gc.duration:PS Scavenge:hist.sum":
            stat_name = "jvm_gc_duration"
            mean = [indicator.data[f"{STAT_PREFIX[i]}.mean"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = mean[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = mean[1]


def calc_stat_for_container_node(node: PlaceNode):
    assert node.kind == PlaceKind.container

    indicator_names = [
        "container.cpu.usage:request_percentage",
        "container.memory.usage:request_percentage",
        "k8s.container.restarts",
    ]

    for name in indicator_names:
        indicator = node.indicators.get(name)
        if indicator is None:
            continue

        assert indicator.df.schema["value"] == pl.Float64
        assert indicator.df["value"].is_not_null().all()

        lf = indicator.df.lazy()
        normal_lf = lf.filter(pl.col("anomal") == 0)
        anomal_lf = lf.filter(pl.col("anomal") == 1)

        df_list = pl.collect_all([normal_lf, anomal_lf])

        value_list = [df["value"].to_numpy() for df in df_list]

        calc_regular_stat(node, indicator, value_list)

        if indicator.name == "container.cpu.usage:request_percentage":
            stat_name = "cpu_usage"
            mean = [indicator.data[f"{STAT_PREFIX[i]}.mean"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = mean[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = mean[1]

        elif indicator.name == "container.memory.usage:request_percentage":
            stat_name = "memory_usage"
            mean = [indicator.data[f"{STAT_PREFIX[i]}.mean"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = mean[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = mean[1]

        elif indicator.name == "k8s.container.restarts":
            stat_name = "restart_count"
            max_value = [indicator.data[f"{STAT_PREFIX[i]}.max"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = max_value[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = max_value[1]


def calc_stat_for_function_node(node: PlaceNode):
    assert node.kind == PlaceKind.function

    indicator_names = [
        "duration",
        "attr.status_code",
        "attr.http.response.status_code",
        "attr.http.request.content_length",
        "attr.http.response.content_length",
    ]

    for name in indicator_names:
        indicator = node.indicators.get(name)
        if indicator is None:
            continue

        assert indicator.df.schema["value"] == pl.Float64
        assert indicator.df["value"].is_not_null().all()

        lf = indicator.df.lazy()
        normal_lf = lf.filter(pl.col("anomal") == 0)
        anomal_lf = lf.filter(pl.col("anomal") == 1)

        df_list = pl.collect_all([normal_lf, anomal_lf])

        value_list = [df["value"].to_numpy() for df in df_list]

        calc_regular_stat(node, indicator, value_list)

        if indicator.name == "duration":
            stat_name = "qpm"
            qpm = [calc_qpm(df_list[i]) for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = qpm[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = qpm[1]

            stat_name = "latency"
            trimmed_mean_p10 = [indicator.data[f"{STAT_PREFIX[i]}.trimmed_mean.p10"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = trimmed_mean_p10[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = trimmed_mean_p10[1]

            stat_name = "latency_p50"
            p50 = [indicator.data[f"{STAT_PREFIX[i]}.quantile.p50"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = p50[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = p50[1]

            stat_name = "latency_p90"
            p90 = [indicator.data[f"{STAT_PREFIX[i]}.quantile.p90"] for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = p90[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = p90[1]

            stat_name = "traces_id_set"
            traces_id_set = [calc_traces_id_set(df_list[i]) for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = traces_id_set[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = traces_id_set[1]

        elif indicator.name == "attr.status_code":
            stat_name = "error_rate"
            error_rate = [calc_error_rate(value_list[i]) for i in range(2)]
            node.data[f"{STAT_PREFIX[0]}.{stat_name}"] = error_rate[0]
            node.data[f"{STAT_PREFIX[1]}.{stat_name}"] = error_rate[1]


def calc_regular_stat(
    node: PlaceNode,
    indicator: Indicator,
    value_list: list[np.ndarray],
):
    stat_functions = [
        ("count", calc_count),
        ("min", calc_min),
        ("max", calc_max),
        ("mean", calc_mean),
    ]

    for p in [0.05, 0.10, 0.25]:
        f = functools.partial(calc_trimmed_mean, proportiontocut=p)
        stat_functions.append((f"trimmed_mean.p{int(p * 100)}", f))

    for q in [0.50, 0.90, 0.95, 0.99]:
        f = functools.partial(calc_quantile, q=q)
        stat_functions.append((f"quantile.p{int(q * 100)}", f))

    for stat_name, stat_func in stat_functions:
        len0 = len(value_list[0])
        len1 = len(value_list[1])

        try:
            if len0 > 0:
                stat0 = stat_func(value_list[0])
            else:
                stat0 = None

            if len1 > 0:
                stat1 = stat_func(value_list[1])
            else:
                stat1 = None
        except Exception as e:
            traceback.print_exc()
            logger.error(f"Failed to calc `{stat_name}` for `{node.uniq_name}|{indicator.name}`: {type(e)}")
            raise

        indicator.data[f"{STAT_PREFIX[0]}.{stat_name}"] = stat0
        indicator.data[f"{STAT_PREFIX[1]}.{stat_name}"] = stat1


def calc_count(value: np.ndarray) -> float:
    return float(len(value))


def calc_min(value: np.ndarray) -> float:
    ans = value.min()
    assert isinstance(ans, float) and not math.isnan(ans), f"ans=`{ans}`"
    return float(ans)


def calc_max(value: np.ndarray) -> float:
    ans = value.max()
    assert isinstance(ans, float) and not math.isnan(ans), f"ans=`{ans}`"
    return float(ans)


def calc_mean(value: np.ndarray) -> float:
    ans = value.mean()
    assert isinstance(ans, float) and not math.isnan(ans), f"ans=`{ans}`"
    return float(ans)


def calc_trimmed_mean(value: np.ndarray, proportiontocut: float) -> float:
    assert 0 < proportiontocut < 1
    ans = scipy.stats.trim_mean(value, proportiontocut)
    assert isinstance(ans, float) and not math.isnan(ans), f"ans=`{ans}`"
    return float(ans)


def calc_quantile(value: np.ndarray, q: float) -> float:
    assert 0 < q < 1
    ans = np.quantile(value, q)
    assert isinstance(ans, float) and not math.isnan(ans), f"ans=`{ans}`"
    return float(ans)


def calc_qpm(duration_df: pl.DataFrame) -> float:
    if len(duration_df) == 0:
        return 0.0

    time_min = duration_df["time"].min()
    time_max = duration_df["time"].max()

    assert isinstance(time_min, datetime.datetime)
    assert isinstance(time_max, datetime.datetime)

    time_in_minutes = (time_max - time_min).total_seconds() / 60.0
    if time_in_minutes == 0:
        return len(duration_df)

    qpm = len(duration_df) / time_in_minutes
    return qpm


def calc_error_rate(status_code: np.ndarray) -> float:
    if len(status_code) == 0:
        return 0.0

    ans = np.mean(status_code == 2)
    assert isinstance(ans, float) and not math.isnan(ans), f"ans=`{ans}`"
    return float(ans)


def calc_traces_id_set(duration_df: pl.DataFrame) -> set[str]:
    return set(duration_df["trace_id"].unique())


def calc_stat_for_calls_edge(sdg: SDG, edge: DepEdge):
    assert edge.kind == DepKind.calls

    src = sdg.get_node_by_id(edge.src_id)
    dst = sdg.get_node_by_id(edge.dst_id)

    for prefix in STAT_PREFIX:
        src_traces_id_set = src.data.get(f"{prefix}.traces_id_set") or set()
        dst_traces_id_set = dst.data.get(f"{prefix}.traces_id_set") or set()

        common_traces_id_set = src_traces_id_set & dst_traces_id_set
        edge.data[f"{prefix}.traces_id_set"] = common_traces_id_set

        if src_traces_id_set:
            forward_call_prob = len(common_traces_id_set) / len(src_traces_id_set)
        else:
            forward_call_prob = 0.0

        if dst_traces_id_set:
            backward_call_prob = len(common_traces_id_set) / len(dst_traces_id_set)
        else:
            backward_call_prob = 0.0

        edge.data[f"{prefix}.forward_call_prob"] = forward_call_prob
        edge.data[f"{prefix}.backward_call_prob"] = backward_call_prob



from typing import Any

import plotly.graph_objects as go
import polars as pl
import streamlit as st
from plotly.subplots import make_subplots

from .utils import get_injection_time_markers


class MetricsVisualizer:
    """Simplified metrics visualizer with Polars acceleration."""

    def __init__(self) -> None:
        """Initialize the metrics visualizer."""
        self.colors = {
            "normal": "#1f77b4",
            "abnormal": "#ff7f0e",
            "injection_start": "#d62728",
            "injection_end": "#2ca02c",
        }

    def _get_time_range(self, df: pl.DataFrame, time_col: str) -> tuple[Any | None, Any | None]:
        """Get the actual time range from the data."""
        try:
            min_time = df[time_col].min()
            max_time = df[time_col].max()
            return min_time, max_time
        except Exception:
            return None, None

    def create_time_series_plot(
        self,
        df: pl.DataFrame,
        metrics: list[str],
        env_data: dict[str, Any],
        title: str = "Time Series Metrics Visualization",
        subplot_height: int = 400,  # 调整默认高度为更合理的400像素
    ) -> go.Figure:
        if len(df) == 0 or not metrics:
            return self._create_empty_figure()

        # Get time column
        time_col = self._get_time_column(df)
        if not time_col:
            st.error("Timestamp column not found")
            return self._create_empty_figure()

        return self._create_long_format_plot(df, metrics, env_data, title, time_col, subplot_height)

    def _create_empty_figure(self) -> go.Figure:
        """Create empty figure with message."""
        fig = go.Figure()
        fig.add_annotation(
            text="No data to display",
            xref="paper",
            yref="paper",
            x=0.5,
            y=0.5,
            showarrow=False,
            font=dict(size=16, color="gray"),
        )
        return fig

    def _create_long_format_plot(
        self,
        df: pl.DataFrame,
        metrics: list[str],
        env_data: dict[str, Any],
        title: str,
        time_col: str,
        subplot_height: int = 400,  # 调整默认高度
    ) -> go.Figure:
        """Create plot for long format data using Polars."""

        # Calculate subplot layout
        import math

        cols = min(3, len(metrics))
        rows = math.ceil(len(metrics) / cols)

        # 根据行数调整间距
        if rows <= 2:
            vertical_spacing = 0.08
        elif rows <= 4:
            vertical_spacing = 0.06
        else:
            vertical_spacing = 0.04

        fig = make_subplots(
            rows=rows,
            cols=cols,
            shared_xaxes=True,
            subplot_titles=metrics,
            vertical_spacing=vertical_spacing,
            horizontal_spacing=0.03,
        )

        # Simple color palette
        colors = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b"]

        for i, metric in enumerate(metrics):
            row = (i // cols) + 1
            col = (i % cols) + 1

            try:
                # Filter data for this metric
                metric_data = (
                    df.filter(pl.col("metric") == metric)
                    .with_columns(pl.col("data_type").fill_null("normal"))
                    .sort([time_col, "service_name"])
                )

                if len(metric_data) == 0:
                    continue

                # Plot each service
                services = metric_data["service_name"].unique().to_list()
                for j, service in enumerate(services):
                    service_data = metric_data.filter(pl.col("service_name") == service)
                    color = colors[j % len(colors)]

                    # Separate normal and abnormal data if available
                    if "data_type" in service_data.columns:
                        normal_data = service_data.filter(pl.col("data_type") == "normal")
                        abnormal_data = service_data.filter(pl.col("data_type") == "abnormal")

                        if len(normal_data) > 0:
                            fig.add_trace(
                                go.Scatter(
                                    x=normal_data[time_col].to_list(),
                                    y=normal_data["value"].to_list(),
                                    mode="lines",
                                    name=f"{service} (Normal)" if i == 0 else None,
                                    line=dict(color=color),
                                    showlegend=(i == 0),
                                    hovertemplate=f"<b>{service}</b><br>%{{x}}<br>%{{y}}<extra></extra>",
                                ),
                                row=row,
                                col=col,
                            )

                        if len(abnormal_data) > 0:
                            fig.add_trace(
                                go.Scatter(
                                    x=abnormal_data[time_col].to_list(),
                                    y=abnormal_data["value"].to_list(),
                                    mode="lines",
                                    name=f"{service} (Abnormal)" if i == 0 else None,
                                    line=dict(color=color, width=3),
                                    showlegend=(i == 0),
                                    hovertemplate=f"<b>{service}</b><br>%{{x}}<br>%{{y}}<extra></extra>",
                                ),
                                row=row,
                                col=col,
                            )
                    else:
                        fig.add_trace(
                            go.Scatter(
                                x=service_data[time_col].to_list(),
                                y=service_data["value"].to_list(),
                                mode="lines",
                                name=service if i == 0 else None,
                                line=dict(color=color),
                                showlegend=(i == 0),
                                hovertemplate=f"<b>{service}</b><br>%{{x}}<br>%{{y}}<extra></extra>",
                            ),
                            row=row,
                            col=col,
                        )

            except Exception as e:
                st.warning(f"Failed to plot metric {metric}: {str(e)}")
                continue

        time_markers = get_injection_time_markers(env_data)
        if time_markers:
            self._add_injection_markers(fig, time_markers, rows, cols)

        min_time, max_time = self._get_time_range(df, time_col)

        ideal_subplot_height = subplot_height
        max_total_height = 3000
        min_subplot_height = 300

        total_height = ideal_subplot_height * rows + 100
        if total_height > max_total_height:
            available_height = max_total_height - 100
            actual_subplot_height = max(min_subplot_height, available_height // rows)
            total_height = actual_subplot_height * rows + 100
        else:
            actual_subplot_height = ideal_subplot_height

        # Update layout
        fig.update_layout(title=title, height=total_height, showlegend=True, hovermode="x unified")

        # Update x-axes with time range for all subplots
        for row_idx in range(1, rows + 1):
            for col_idx in range(1, cols + 1):
                fig.update_xaxes(
                    title_text="Timestamp" if row_idx == rows else None,
                    range=[min_time, max_time] if min_time and max_time else None,
                    row=row_idx,
                    col=col_idx,
                )

        return fig

    def _get_time_column(self, df: pl.DataFrame) -> str | None:
        """Get timestamp column name."""
        time_columns = ["timestamp", "time", "datetime", "ts"]

        for col in time_columns:
            if col in df.columns:
                return col

        # Look for columns with 'time' in name
        for col in df.columns:
            if "time" in col.lower():
                col_dtype = str(df[col].dtype)
                if any(dtype in col_dtype.lower() for dtype in ["int", "float", "datetime"]):
                    return col

        return None

    def _add_injection_markers(self, fig: go.Figure, time_markers: dict[str, Any], rows: int, cols: int):
        """Add fault injection markers to all subplots."""

        if not time_markers:
            return

        # Add start marker
        if "abnormal_start" in time_markers:
            timestamp = time_markers["abnormal_start"]

            for row in range(1, rows + 1):
                for col in range(1, cols + 1):
                    fig.add_shape(
                        type="line",
                        x0=timestamp,
                        x1=timestamp,
                        y0=0,
                        y1=1,
                        yref="paper",
                        line=dict(color=self.colors["injection_start"], dash="dash", width=2),
                        row=row,
                        col=col,
                    )

            fig.add_annotation(
                x=timestamp,
                y=0.95,
                yref="paper",
                text="Fault start",
                showarrow=False,
                font=dict(color=self.colors["injection_start"], size=10),
                bgcolor="white",
                bordercolor=self.colors["injection_start"],
                borderwidth=1,
            )

        # Add end marker
        if "abnormal_end" in time_markers:
            timestamp = time_markers["abnormal_end"]

            for row in range(1, rows + 1):
                for col in range(1, cols + 1):
                    fig.add_shape(
                        type="line",
                        x0=timestamp,
                        x1=timestamp,
                        y0=0,
                        y1=1,
                        yref="paper",
                        line=dict(color=self.colors["injection_end"], dash="dash", width=2),
                        row=row,
                        col=col,
                    )

            fig.add_annotation(
                x=timestamp,
                y=0.90,
                yref="paper",
                text="Fault end",
                showarrow=False,
                font=dict(color=self.colors["injection_end"], size=10),
                bgcolor="white",
                bordercolor=self.colors["injection_end"],
                borderwidth=1,
            )

    def _add_single_plot_markers(self, fig: go.Figure, time_markers: dict[str, Any]):
        """Add fault injection markers to single plot."""

        if not time_markers:
            return

        # Add start marker
        if "abnormal_start" in time_markers:
            timestamp = time_markers["abnormal_start"]
            fig.add_shape(
                type="line",
                x0=timestamp,
                x1=timestamp,
                y0=0,
                y1=1,
                yref="paper",
                line=dict(color=self.colors["injection_start"], dash="dash", width=2),
            )
            fig.add_annotation(
                x=timestamp,
                y=0.95,
                yref="paper",
                text="Fault start",
                showarrow=False,
                font=dict(color=self.colors["injection_start"], size=10),
                bgcolor="white",
                bordercolor=self.colors["injection_start"],
                borderwidth=1,
            )

        # Add end marker
        if "abnormal_end" in time_markers:
            timestamp = time_markers["abnormal_end"]
            fig.add_shape(
                type="line",
                x0=timestamp,
                x1=timestamp,
                y0=0,
                y1=1,
                yref="paper",
                line=dict(color=self.colors["injection_end"], dash="dash", width=2),
            )
            fig.add_annotation(
                x=timestamp,
                y=0.90,
                yref="paper",
                text="Fault end",
                showarrow=False,
                font=dict(color=self.colors["injection_end"], size=10),
                bgcolor="white",
                bordercolor=self.colors["injection_end"],
                borderwidth=1,
            )



import functools
import time
from collections import defaultdict
from collections.abc import Callable, Generator
from contextlib import contextmanager
from typing import Any, TypeVar

F = TypeVar("F", bound=Callable[..., Any])


class ProfilerResult:
    def __init__(self):
        self.timings: dict[str, list[float]] = defaultdict(list)
        self.call_counts: dict[str, int] = defaultdict(int)

    def record(self, name: str, duration: float) -> None:
        self.timings[name].append(duration)
        self.call_counts[name] += 1

    def get_stats(self) -> dict[str, dict[str, float]]:
        stats = {}
        for name, times in self.timings.items():
            stats[name] = {
                "count": self.call_counts[name],
                "total": sum(times),
                "avg": sum(times) / len(times),
                "min": min(times),
                "max": max(times),
            }
        return stats

    def print_stats(self, sort_by: str = "total") -> None:
        stats = self.get_stats()

        if not stats:
            print("No profiling data available.")
            return

        sorted_stats = sorted(stats.items(), key=lambda x: x[1][sort_by], reverse=True)

        print(f"\n{'Function':<40} {'Count':<8} {'Total(s)':<10} {'Avg(s)':<10} {'Min(s)':<10} {'Max(s)':<10}")
        print("-" * 88)

        for name, data in sorted_stats:
            print(
                f"{name:<40} {data['count']:<8} {data['total']:<10.4f} {data['avg']:<10.4f} "
                f"{data['min']:<10.4f} {data['max']:<10.4f}"
            )

        print("-" * 88)
        total_time = sum(data["total"] for data in stats.values())
        print(f"{'Total execution time:':<40} {total_time:.4f}s")


class FunctionProfiler:
    def __init__(self):
        self.result = ProfilerResult()
        self.enabled = True

    def enable(self) -> None:
        self.enabled = True

    def disable(self) -> None:
        self.enabled = False

    def clear(self) -> None:
        self.result = ProfilerResult()

    @contextmanager
    def profile(self, name: str) -> Generator[None, None, None]:
        if not self.enabled:
            yield
            return

        start_time = time.perf_counter()
        try:
            yield
        finally:
            end_time = time.perf_counter()
            duration = end_time - start_time
            self.result.record(name, duration)

    def profile_function(self, name: str | None = None) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            func_name = name or f"{func.__module__}.{func.__qualname__}"

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if not self.enabled:
                    return func(*args, **kwargs)

                with self.profile(func_name):
                    return func(*args, **kwargs)

            return wrapper  # type: ignore

        return decorator

    def profile_method_calls(self, obj: Any, method_names: list[str]) -> None:
        for method_name in method_names:
            if hasattr(obj, method_name):
                original_method = getattr(obj, method_name)
                profiled_method = self.profile_function(f"{obj.__class__.__name__}.{method_name}")(original_method)
                setattr(obj, method_name, profiled_method)


global_profiler = FunctionProfiler()


def profile(name: str | None = None):
    return global_profiler.profile_function(name)


@contextmanager
def profile_block(name: str):
    with global_profiler.profile(name):
        yield


def get_profiler_stats():
    return global_profiler.result.get_stats()


def print_profiler_stats(sort_by: str = "total"):
    global_profiler.result.print_stats(sort_by)


def clear_profiler():
    global_profiler.clear()


def enable_profiler():
    global_profiler.enable()


def disable_profiler():
    global_profiler.disable()



import functools
from typing import Annotated

import typer

from ..datasets.spec import get_datapack_folder, get_dataset_list
from ..logging import logger, timeit
from ..samplers.experiments.batch import run_sampler_batch
from ..samplers.experiments.report import generate_sampler_perf_report
from ..samplers.experiments.single import run_sampler_single
from ..samplers.metrics_sli import generate_metrics_sli
from ..samplers.spec import SamplingMode, global_sampler_registry
from ..utils.fmap import fmap_processpool

app = typer.Typer()


@app.command()
def show_samplers():
    """Show available sampler algorithms."""
    registry = global_sampler_registry()
    logger.info(f"Available samplers ({len(registry)}):")
    for name in registry:
        logger.info(f"    {name}")


@app.command()
@timeit()
def single(
    sampler: str,
    dataset: str,
    datapack: str,
    sampling_rate: float = 0.1,
    mode: SamplingMode = SamplingMode.OFFLINE,
    clear: bool = False,
    skip_finished: bool = True,
):
    """Run a single sampler on a datapack."""
    run_sampler_single(
        sampler=sampler,
        dataset=dataset,
        datapack=datapack,
        sampling_rate=sampling_rate,
        mode=mode,
        clear=clear,
        skip_finished=skip_finished,
    )


@app.command()
@timeit()
def batch(
    samplers: Annotated[list[str], typer.Option("-s", "--sampler")],
    datasets: Annotated[list[str], typer.Option("-d", "--dataset")],
    sampling_rates: Annotated[list[float], typer.Option("-r", "--rate")] = [0.1],
    modes: Annotated[list[SamplingMode], typer.Option("-m", "--mode")] = [SamplingMode.OFFLINE],
    sample_datapacks: int | None = None,
    clear: bool = False,
    skip_finished: bool = True,
    use_cpus: int | None = None,
    ignore_exceptions: bool = True,
):
    """Run multiple samplers on multiple datasets."""
    run_sampler_batch(
        samplers=samplers,
        datasets=datasets,
        sampling_rates=sampling_rates,
        modes=modes,
        sample_datapacks=sample_datapacks,
        clear=clear,
        skip_finished=skip_finished,
        use_cpus=use_cpus,
        ignore_exceptions=ignore_exceptions,
    )


@app.command()
@timeit()
def perf_report(
    datasets: Annotated[list[str], typer.Option("-d", "--dataset")],
    samplers: Annotated[list[str], typer.Option("-s", "--sampler")] | None = None,
    sampling_rates: Annotated[list[float], typer.Option("-r", "--rate")] | None = None,
    modes: Annotated[list[SamplingMode], typer.Option("-m", "--mode")] | None = None,
    warn_missing: bool = False,
):
    """
    Generate performance report for samplers.

    By default, auto-detects all available samplers, sampling rates, and modes.
    Use filters to specify particular configurations:

    Examples:
        # Report all configurations for datasets
        sample perf-report -d dataset1 -d dataset2

        # Report specific sampler only
        sample perf-report -d dataset1 -s random

        # Report specific rates and modes
        sample perf-report -d dataset1 -r 0.1 -r 0.2 -m offline

        # Report specific combinations
        sample perf-report -d dataset1 -s random -s my_sampler -r 0.1 -m offline -m online
    """
    if not datasets:
        logger.error("At least one dataset must be specified with -d/--dataset")
        raise typer.Exit(1)

    # Show what will be processed
    if samplers is None:
        logger.info("Auto-detecting available samplers...")
    else:
        logger.info(f"Filtering samplers: {samplers}")

    if sampling_rates is None:
        logger.info("Auto-detecting available sampling rates...")
    else:
        logger.info(f"Filtering sampling rates: {sampling_rates}")

    if modes is None:
        logger.info("Auto-detecting available sampling modes...")
    else:
        logger.info(f"Filtering sampling modes: {[m.value for m in modes]}")

    generate_sampler_perf_report(
        datasets=datasets,
        samplers=samplers,
        sampling_rates=sampling_rates,
        modes=modes,
        warn_missing=warn_missing,
    )


@app.command()
@timeit()
def generate_sli_metrics(
    dataset: str,
    datapack: str | None = None,
):
    """Generate metrics_sli.parquet for dataset or specific datapack."""
    if datapack is not None:
        # Generate for specific datapack
        input_folder = get_datapack_folder(dataset, datapack)
        generate_metrics_sli(input_folder)
    else:
        # Generate for all datapacks in dataset
        from ..datasets.spec import get_datapack_list

        datapacks = get_datapack_list(dataset)
        logger.info(f"Generating metrics_sli.parquet for {len(datapacks)} datapacks in {dataset}")
        tasks = []

        for dp in datapacks:
            input_folder = get_datapack_folder(dataset, dp)
            tasks.append(functools.partial(generate_metrics_sli, input_folder))

        fmap_processpool(tasks, parallel=64)



from pathlib import Path

from ..config import get_config


def get_output_folder(dataset: str, datapack: str, algorithm: str) -> Path:
    config = get_config()
    return config.output / "data" / dataset / datapack / algorithm


def get_output_meta_folder(dataset: str) -> Path:
    config = get_config()
    return config.output / "meta" / dataset



import multiprocessing

import typer
from tqdm.auto import tqdm

from ..algorithms.spec import global_algorithm_registry
from ..logging import logger, timeit
from ..utils.env import getenv_bool

app = typer.Typer(pretty_exceptions_show_locals=False)


@app.callback()
def _():
    logger.remove()
    logger.add(
        lambda msg: tqdm.write(msg, end=""),
        colorize=getenv_bool("LOGURU_COLORIZE", default=True),
        enqueue=True,
        context=multiprocessing.get_context("spawn"),
    )


def main(*, enable_builtin_algorithms: bool = True) -> None:
    from . import container, eval, online, sample, sdg, self_, tools

    app.add_typer(self_.app, name="self")
    app.add_typer(tools.app, name="tools")
    app.add_typer(online.app, name="online")
    app.add_typer(sdg.app, name="sdg")
    app.add_typer(eval.app, name="eval")
    app.add_typer(sample.app, name="sample")
    app.add_typer(container.app, name="container")
    if enable_builtin_algorithms:
        register_builtin_algorithms()

    app()


def register_builtin_algorithms():
    from ..algorithms.random_ import Random
    from ..algorithms.rcaeval.baro import Baro
    from ..algorithms.rcaeval.nsigma import NSigma
    from ..algorithms.traceback.A7 import TraceBackA7
    from ..algorithms.traceback.A8 import TraceBackA8

    getters = {
        "random": Random,
        "traceback-A7": TraceBackA7,
        "traceback-A8": TraceBackA8,
        "baro": Baro,
        "nsigma": NSigma,
    }

    registry = global_algorithm_registry()
    for name, getter in getters.items():
        registry[name] = getter



import polars as pl

from ..defintion import SDG, DepEdge, DepKind, Indicator, PlaceKind, PlaceNode


def calc_metric_min_max(df: pl.DataFrame) -> tuple[float, float]:
    col = pl.col("value")
    df = df.select(min=col.min(), max=col.max())
    min_value, max_value = df.row(0)
    assert isinstance(min_value, float)
    assert isinstance(max_value, float)
    return min_value, max_value


def is_constant_metric(df: pl.DataFrame) -> bool:
    min_value, max_value = calc_metric_min_max(df)
    return (max_value - min_value) < 1e-8


def replace_enum_values(col_name: str, values: list[str], *, start: int) -> pl.Expr:
    return (
        pl.col(col_name)
        .cast(pl.Enum(values))
        .replace_strict({value: i for i, value in enumerate(values, start=start)})
        .cast(pl.Float64)
    )


def add_node_opt(sdg: SDG, kind: PlaceKind, self_name: str | None) -> PlaceNode | None:
    if self_name is None:
        return None
    return sdg.add_node(PlaceNode(kind=kind, self_name=self_name), strict=False)



import json
import math
import os
import random
import re
from collections import defaultdict
from pathlib import Path
from typing import Any

import networkx as nx
import polars as pl

from ..datasets.spec import get_datapack_folder
from ..logging import logger, timeit
from ..sources.convert import link_subset
from .spec import DatasetAnalyzer, build_service_graph

DATAPACK_PATTERN = (
    r"(ts|ts\d)-(mysql|ts-rabbitmq|ts-ui-dashboard|ts-\w+-service|ts-\w+-\w+-service|ts-\w+-\w+-\w+-service)-(.+)-[^-]+"
)


def rcabench_get_service_name(datapack_name: str) -> str:
    m = re.match(DATAPACK_PATTERN, datapack_name)
    assert m is not None, f"Invalid datapack name: `{datapack_name}`"
    service_name: str = m.group(2)
    return service_name


FAULT_TYPES: list[str] = [
    "PodKill",
    "PodFailure",
    "ContainerKill",
    "MemoryStress",
    "CPUStress",
    "HTTPRequestAbort",
    "HTTPResponseAbort",
    "HTTPRequestDelay",
    "HTTPResponseDelay",
    "HTTPResponseReplaceBody",
    "HTTPResponsePatchBody",
    "HTTPRequestReplacePath",
    "HTTPRequestReplaceMethod",
    "HTTPResponseReplaceCode",
    "DNSError",
    "DNSRandom",
    "TimeSkew",
    "NetworkDelay",
    "NetworkLoss",
    "NetworkDuplicate",
    "NetworkCorrupt",
    "NetworkBandwidth",
    "NetworkPartition",
    "JVMLatency",
    "JVMReturn",
    "JVMException",
    "JVMGarbageCollector",
    "JVMCPUStress",
    "JVMMemoryStress",
    "JVMMySQLLatency",
    "JVMMySQLException",
]


def get_parent_resource_from_pod_name(
    pod_name: str,
) -> tuple[str | None, str | None, str | None]:
    """
    Parse parent resource from Pod name (Deployment + ReplicaSet or StatefulSet/DaemonSet)

    Supported parent resource types:
    - Deployment Pods: <deployment-name>-<replicaset-hash>-<pod-hash>
        → Returns ("Deployment", deployment_name, replicaset_name)
    - StatefulSet Pods: <statefulset-name>-<ordinal>
        → Returns ("StatefulSet", statefulset_name, None)
    - DaemonSet Pods: <daemonset-name>-<pod-hash>
        → Returns ("DaemonSet", daemonset_name, None)
    - Other cases return (None, None, None)

    Args:
        podname (str): Pod name

    Returns:
        tuple: (parent_type, parent_name, replicaset_name_if_applicable)
    """
    # Deployment Pod format: <deployment-name>-<replicaset-hash>-<pod-hash>
    # Example: nginx-deployment-5c689d88bb-q7zvf
    deployment_pattern = r"^(?P<deploy>.+?)-(?P<rs_hash>[a-z0-9]{5,10})-(?P<pod_hash>[a-z0-9]{5})$"
    match = re.fullmatch(deployment_pattern, pod_name)
    if match:
        deployment_name = match.group("deploy")
        replicaset_name = f"{deployment_name}-{match.group('rs_hash')}"
        return ("Deployment", deployment_name, replicaset_name)

    # StatefulSet Pod format: <statefulset-name>-<ordinal>
    # Example: web-0, mysql-1
    statefulset_pattern = r"^(?P<sts>.+)-(\d+)$"
    match = re.fullmatch(statefulset_pattern, pod_name)
    if match:
        return ("StatefulSet", match.group("sts"), None)

    # DaemonSet Pod format: <daemonset-name>-<pod-hash>
    # Example: fluentd-elasticsearch-abcde
    daemonset_pattern = r"^(?P<ds>.+)-([a-z0-9]{5})$"
    match = re.fullmatch(daemonset_pattern, pod_name)
    if match:
        return ("DaemonSet", match.group("ds"), None)

    # Other cases (like bare Pod or unknown format)
    return (None, None, None)


HTTP_REPLACE_METHODS: list[str] = [
    "GET",
    "POST",
    "PUT",
    "DELETE",
    "HEAD",
    "OPTIONS",
    "PATCH",
]

HTTP_REPLACE_BODY_TYPE: dict[int, str] = {
    0: "empty",
    1: "random",
}

JVM_MEM_TYPE: dict[int, str] = {
    1: "heap",
    2: "stack",
}

JVM_RETURN_TYPE: dict[int, str] = {
    1: "String",
    2: "Int",
}

JVM_RETURN_VALUE_OPT: dict[int, str] = {
    0: "Default",
    1: "Random",
}


def rcabench_fix_injection(injection: dict[str, Any]) -> None:
    injection["fault_type"] = FAULT_TYPES[injection["fault_type"]]

    injection["engine_config"] = json.loads(injection["engine_config"])

    display_config: dict[str, Any] = json.loads(injection["display_config"])
    rcabench_fix_injection_display_config(display_config)
    injection["display_config"] = display_config


def rcabench_fix_injection_display_config(display_config: dict[str, Any]) -> None:
    if (replace_method := display_config.get("replace_method")) is not None:
        if isinstance(replace_method, int):
            display_config["replace_method"] = HTTP_REPLACE_METHODS[replace_method]
        elif isinstance(replace_method, str):
            pass
        else:
            raise ValueError(f"Invalid replace_method type: {type(replace_method)}. Expected int or str.")

    replacements = [
        ("body_type", HTTP_REPLACE_BODY_TYPE),
        ("mem_type", JVM_MEM_TYPE),
        ("return_type", JVM_RETURN_TYPE),
        ("return_value_opt", JVM_RETURN_VALUE_OPT),
    ]

    for k, d in replacements:
        v = display_config.get(k)
        if v is None:
            continue
        display_config[k] = d[v]


@timeit(log_args={"train_ratio"})
def rcabench_split_train_test(
    datapacks: list[str],
    train_ratio: float,
    previous_datapacks: list[str],
    datapack_limit: int = 0,
):
    assert len(datapacks) > 0, "Datapacks list cannot be empty."
    assert 0 < train_ratio < 1, "Ratio must be between 0 and 1."
    assert datapack_limit <= len(datapacks), "Datapack limit must be less than or equal to the number of datapacks."

    prev_datapacks = set(previous_datapacks)
    additional_datapacks = set(datapacks) - prev_datapacks

    group_by_service: defaultdict[str, list[str]] = defaultdict(list)
    for datapack in additional_datapacks:
        service_name = rcabench_get_service_name(datapack)
        group_by_service[service_name].append(datapack)

    min_group_size = min(len(v) for v in group_by_service.values())
    logger.debug("min_group_size: {}", min_group_size)

    threshold = min_group_size
    while True:
        train_total = 0
        for service_datapacks in group_by_service.values():
            num_train = math.ceil(len(service_datapacks) * train_ratio)
            num_train = min(num_train, threshold)
            train_total += num_train

        target = len(additional_datapacks) * train_ratio

        logger.debug("threshold={} (train_total={}, target={})", threshold, train_total, target)

        if train_total >= target:
            break

        threshold += 1

    train_datapacks: list[str] = []
    test_datapacks: list[str] = []

    for service_name, service_datapacks in group_by_service.items():
        random.shuffle(service_datapacks)

        num_train = math.ceil(len(service_datapacks) * train_ratio)
        num_train = min(num_train, threshold)

        train_datapacks.extend(service_datapacks[:num_train])
        test_datapacks.extend(service_datapacks[num_train:])

    total_selected = len(train_datapacks) + len(test_datapacks)

    if total_selected > datapack_limit:
        target_train = int(datapack_limit * train_ratio)
        target_test = datapack_limit - target_train

        train_datapacks = train_datapacks[:target_train]
        test_datapacks = test_datapacks[:target_test]

        logger.info("Adjusted to datapack_limit: train={}, test={}", len(train_datapacks), len(test_datapacks))

    logger.info(
        "Final dataset: train={} datapacks, test={} datapacks, total={}",
        len(train_datapacks),
        len(test_datapacks),
        len(train_datapacks) + len(test_datapacks),
    )

    return train_datapacks, test_datapacks


def valid(path: Path) -> tuple[Path, bool]:
    path_obj = path

    if not path_obj.exists() or not path_obj.is_dir():
        logger.debug("Path does not exist or is not a directory: {}", path)
        return path, False

    # Check cache files first
    valid_cache = path_obj / ".valid"
    invalid_cache = path_obj / ".invalid"

    if valid_cache.exists():
        return path, True
    elif invalid_cache.exists():
        return path, False

    required_files = [
        # Parquet files
        "abnormal_logs.parquet",
        "abnormal_metrics_sum.parquet",
        "abnormal_metrics_histogram.parquet",
        "abnormal_trace_id_ts.parquet",
        "abnormal_metrics.parquet",
        "abnormal_traces.parquet",
        "normal_metrics_histogram.parquet",
        "normal_trace_id_ts.parquet",
        "normal_logs.parquet",
        "normal_metrics.parquet",
        "normal_metrics_sum.parquet",
        "normal_traces.parquet",
        # JSON files
        "injection.json",
        "k8s.json",
        "env.json",
    ]

    for filename in required_files:
        file_path = path_obj / filename

        if not file_path.exists():
            logger.debug("Missing required file: {}", file_path)
            invalid_f = path_obj / ".invalid"
            invalid_f.touch()
            return path, False

        if file_path.stat().st_size == 0:
            logger.debug("Empty file: {}", file_path)
            invalid_f = path_obj / ".invalid"
            invalid_f.touch()
            return path, False

        if filename.endswith(".json"):
            try:
                with open(file_path, encoding="utf-8") as f:
                    json.load(f)
            except (json.JSONDecodeError, UnicodeDecodeError) as e:
                logger.debug("Invalid JSON file {}: {}", file_path, e)
                invalid_f = path_obj / ".invalid"
                invalid_f.touch()
                return path, False

        elif filename.endswith(".parquet"):
            try:
                df = pl.read_parquet(file_path)
                row_count = df.height

                if row_count == 0:
                    logger.debug("Parquet file has no data rows: {}", file_path)
                    invalid_f = path_obj / ".invalid"
                    invalid_f.touch()
                    return path, False

            except Exception as e:
                logger.debug("Failed to read Parquet file {}: {}", file_path, e)
                invalid_f = path_obj / ".invalid"
                invalid_f.touch()
                return path, False

    # All validation passed, create valid cache file
    valid_f = path_obj / ".valid"
    valid_f.touch()
    return path, True


class RCABenchAnalyzerLoader(DatasetAnalyzer):
    # Golden signal metrics definition (SRE four golden signals: latency, traffic, errors, saturation)

    _GOLDEN_SIGNAL_METRICS = {
        "latency": [
            "http.client.request.duration",
            "http.server.request.duration",
            "db.client.connections.use_time",
            "db.client.connections.create_time",
            "db.client.connections.wait_time",
            "jvm.gc.duration",
        ],
        "traffic": [
            "hubble_flows_processed_total",
            "processedSpans",
            "processedLogs",
            "hubble_icmp_total",
            "hubble_port_distribution_total",
            "hubble_tcp_flags_total",
            "k8s.pod.network.io",
        ],
        "error": [
            "hubble_drop_total",
            "k8s.pod.network.errors",
            "k8s.container.restarts",
        ],
        "saturation": [
            "container.cpu.usage",
            "k8s.pod.cpu.usage",
            "k8s.pod.cpu_limit_utilization",
            "k8s.pod.cpu.node.utilization",
            "jvm.cpu.recent_utilization",
            "jvm.system.cpu.utilization",
            "jvm.system.cpu.load_1m",
            "container.memory.usage",
            "k8s.pod.memory.usage",
            "k8s.pod.memory_limit_utilization",
            "k8s.pod.memory.node.utilization",
            "container.memory.working_set",
            "k8s.pod.memory.working_set",
            "jvm.memory.used",
            "container.filesystem.usage",
            "k8s.pod.filesystem.usage",
            "queueSize",
        ],
    }

    def __init__(self, datapack: str):
        super().__init__(datapack)
        self.files: dict[str, Any] = self._load_datapack_files()

    def _get_datapack_folder(self) -> Path | None:
        return Path("data") / "rcabench_dataset" / self.datapack / "converted"

    def _load_datapack_files(self) -> dict[str, Any]:
        folder = self._get_datapack_folder()
        assert folder is not None, "datapack folder must exist"
        files: dict[str, Any] = {}

        for file_type in [
            "traces",
            "logs",
            "metrics",
            "metrics_sum",
            "metrics_histogram",
        ]:
            normal_file = folder / f"normal_{file_type}.parquet"
            abnormal_file = folder / f"abnormal_{file_type}.parquet"
            if normal_file.exists():
                lf = pl.scan_parquet(normal_file)
                lf = lf.sort("time")
                files[f"normal_{file_type}"] = lf
            if abnormal_file.exists():
                lf = pl.scan_parquet(abnormal_file)
                lf = lf.sort("time")
                files[f"abnormal_{file_type}"] = lf

        for json_file in ["env.json", "injection.json"]:
            json_path = folder / json_file
            if json_path.exists():
                with open(json_path) as f:
                    files[json_file.replace(".json", "")] = json.load(f)

        conclusion_file = folder / "conclusion.parquet"
        if conclusion_file.exists():
            lf = pl.scan_parquet(conclusion_file)
            files["conclusion"] = lf

        return files

    def get_traces(self, abnormal: bool = False) -> pl.LazyFrame | None:
        key = "abnormal_traces" if abnormal else "normal_traces"
        return self.files.get(key)

    def get_metrics(self, abnormal: bool = False) -> pl.LazyFrame | None:
        key = "abnormal_metrics" if abnormal else "normal_metrics"
        return self.files.get(key)

    def get_logs(self, abnormal: bool = False) -> pl.LazyFrame | None:
        key = "abnormal_logs" if abnormal else "normal_logs"
        return self.files.get(key)

    def get_conclusion(self) -> pl.LazyFrame | None:
        return self.files.get("conclusion")

    def get_service_dependency_graph(self) -> nx.DiGraph:
        folder = self._get_datapack_folder()
        assert folder is not None, "datapack folder must exist"

        normal_traces = pl.scan_parquet(folder / "normal_traces.parquet")
        anomal_traces = pl.scan_parquet(folder / "abnormal_traces.parquet")
        traces = pl.concat([normal_traces, anomal_traces])

        return build_service_graph(traces)

    def get_all_services(self) -> list[str]:
        services = set()
        for key in [
            "normal_traces",
            "abnormal_traces",
            "normal_metrics",
            "abnormal_metrics",
        ]:
            lf = self.files.get(key)
            if lf is not None and "service_name" in lf.collect_schema():
                services.update(lf.select("service_name").unique().collect()["service_name"].to_list())

        services.discard("loadgenerator-service")
        services.discard("loadgenerator")
        services.discard("")
        return list(services)

    def get_service_metrics(self, service_name: str, abnormal: bool = False) -> dict[str, list[float]]:
        metrics_lf = self.get_metrics(abnormal=abnormal)
        traces_lf = self.get_traces(abnormal=abnormal)
        logs_lf = self.get_logs(abnormal=abnormal)
        assert metrics_lf is not None and traces_lf is not None and logs_lf is not None

        return self._extract_service_metrics(metrics_lf, traces_lf, logs_lf, service_name)

    def get_root_services(self) -> list[str]:
        injection = self.files.get("injection", {})
        assert isinstance(injection, dict), "injection must be a dictionary"
        if not injection:
            return []
        ground_truth = injection.get("ground_truth", {})
        root_services = ground_truth.get("service", [])
        return root_services

    def get_entry_service(self) -> str | None:
        return "loadgenerator"

    def _extract_service_metrics(
        self, metrics_lf: pl.LazyFrame, traces_lf: pl.LazyFrame, logs_lf: pl.LazyFrame, service_name: str
    ) -> dict[str, list[float]]:
        assert isinstance(metrics_lf, pl.LazyFrame), "metrics_lf must be a polars LazyFrame"
        assert isinstance(service_name, str) and service_name.strip(), "service_name must be a non-empty string"

        schema = metrics_lf.collect_schema()
        assert "service_name" in schema, "metrics_lf must have service_name column"
        assert "metric" in schema, "metrics_lf must have metric column"
        assert "value" in schema, "metrics_lf must have value column"

        service_metrics = (
            metrics_lf.filter(pl.col("service_name") == service_name)
            .group_by("metric")
            .agg(pl.col("value").alias("values"))
            .collect()
        )

        metrics_dict = {}
        for row in service_metrics.iter_rows(named=True):
            metric_name = row["metric"]
            values = row["values"]
            if not self._is_golden_signal_metric(metric_name):
                continue
            for value in values:
                assert isinstance(value, (int, float)), f"metric value must be numeric, got {type(value)}"
            metrics_dict[metric_name] = values

        error_rate_values = self._calculate_http_error_rate(traces_lf, service_name)
        if error_rate_values:
            metrics_dict["http.response.error_rate"] = error_rate_values

        # Calculate log-based error rate
        log_error_rate_values = self._calculate_log_error_rate(logs_lf, service_name)
        if log_error_rate_values:
            metrics_dict["log.error_rate"] = log_error_rate_values

        # Calculate duration-based latency metrics from traces
        duration_metrics = self._calculate_duration_metrics(traces_lf, service_name)
        metrics_dict.update(duration_metrics)

        return metrics_dict

    def _calculate_http_error_rate(self, traces_lf: pl.LazyFrame, service_name: str) -> list[float] | None:
        service_traces = traces_lf.filter(pl.col("service_name") == service_name)

        # Check if traces have HTTP status codes
        traces_schema = traces_lf.collect_schema()
        if "attr.http.response.status_code" not in traces_schema:
            return None

        # Filter traces that have HTTP status codes
        http_traces = service_traces.filter(pl.col("attr.http.response.status_code").is_not_null())

        if http_traces.collect().height == 0:
            return None

        error_rates = (
            http_traces.with_columns(
                [
                    (pl.col("time").dt.truncate("10s")).alias("time_window"),
                    (pl.col("attr.http.response.status_code") >= 400).alias("is_error"),
                ]
            )
            .group_by("time_window")
            .agg(
                [
                    pl.col("is_error").sum().alias("error_count"),
                    pl.col("is_error").count().alias("total_count"),
                ]
            )
            .with_columns(
                [(pl.col("error_count").cast(pl.Float64) / pl.col("total_count") * 100.0).alias("error_rate")]
            )
            .select("error_rate")
            .collect()
        )

        if error_rates.height == 0:
            return None

        error_rate_values = error_rates["error_rate"].to_list()
        error_rate_values = [v for v in error_rate_values if v is not None]

        return error_rate_values if error_rate_values else None

    def _calculate_log_error_rate(self, logs_lf: pl.LazyFrame, service_name: str) -> list[float] | None:
        service_logs = logs_lf.filter(pl.col("service_name") == service_name)

        # Check if logs have level column
        logs_schema = logs_lf.collect_schema()
        if "level" not in logs_schema:
            return None

        # Filter logs that have valid level values
        valid_logs = service_logs.filter(pl.col("level").is_not_null())

        if valid_logs.collect().height == 0:
            return None

        error_rates = (
            valid_logs.with_columns(
                [
                    (pl.col("time").dt.truncate("10s")).alias("time_window"),
                    (pl.col("level").is_in(["WARN", "ERROR"])).alias("is_error"),
                ]
            )
            .group_by("time_window")
            .agg(
                [
                    pl.col("is_error").sum().alias("error_count"),
                    pl.col("is_error").count().alias("total_count"),
                ]
            )
            .with_columns([(pl.col("error_count").cast(pl.Float64) / pl.col("total_count")).alias("error_rate")])
            .select("error_rate")
            .collect()
        )

        if error_rates.height == 0:
            return None

        error_rate_values = error_rates["error_rate"].to_list()
        return error_rate_values

    def _is_golden_signal_metric(self, metric_name: str) -> bool:
        for metrics_list in self._GOLDEN_SIGNAL_METRICS.values():
            if metric_name in metrics_list:
                return True
            for metric in metrics_list:
                if metric in metric_name:
                    return True
        return False

    def _calculate_duration_metrics(self, traces_lf: pl.LazyFrame, service_name: str) -> dict[str, list[float]]:
        service_traces = traces_lf.filter(pl.col("service_name") == service_name)

        # Check if traces have duration column
        traces_schema = traces_lf.collect_schema()
        if "duration" not in traces_schema:
            return {}

        # Filter traces that have valid duration values (> 0)
        valid_duration_traces = service_traces.filter(pl.col("duration").is_not_null() & (pl.col("duration") > 0))

        # Check if there are any valid traces
        if valid_duration_traces.collect().height == 0:
            return {}

        duration_metrics = {}

        # Calculate time-window based duration metrics (per minute)
        time_window_durations = (
            valid_duration_traces.with_columns(
                [
                    (pl.col("time").dt.truncate("10s")).alias("time_window"),
                    (pl.col("duration") / 1_000_000.0).alias("duration_ms"),  # Convert nanoseconds to milliseconds
                ]
            )
            .group_by("time_window")
            .agg(
                [
                    pl.col("duration_ms").mean().alias("mean_duration"),
                ]
            )
            .collect()
        )

        if time_window_durations.height > 0:
            # Extract time series values for mean
            mean_duration_values = time_window_durations["mean_duration"].to_list()
            mean_duration_values = [v for v in mean_duration_values if v is not None]
            if mean_duration_values:
                duration_metrics["http.server.request.duration"] = mean_duration_values

        return duration_metrics
